"""
Convenience functions to add tracking correction weights.

Felix Metzner 2019
"""

import pandas as pd

from typing import Dict

from analysistools.tracking_correction.tracking_correction import TrackingCorrection

from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection

__all__ = [
    "TrackingCorrection",
    "TrackingEffCorrector",
]


class TrackingEffCorrector:
    tracking_systematic_name = "__tracking_systematic__"
    slow_pi_corr_col_name = "__weight_slow_pi_ratio__"
    slow_pi_corr_var_prefix = "__weight_slow_pi_var_"

    def __init__(self, consider_pi0_for_sys: bool = True) -> None:
        self.exp_column_name = "expNum"

        self.tracking_corrector = TrackingCorrection(
            exp_column_name=self.exp_column_name,
            consider_pi0_for_sys=consider_pi0_for_sys,
        )

        for name, track_info in self.get_track_infos().items():
            self.tracking_corrector.add_track(name=name, **track_info)

    def apply_tracking_correction(self, sample: pd.DataFrame) -> None:
        for col in [c for infos in self.get_track_infos().values() for c in infos.values()]:
            assert col in sample.columns, f"{col}"

        assert self.tracking_systematic_name not in sample.columns

        high_pt_corr, _, slow_pi_corr, slow_pi_col_names = self.tracking_corrector.get_corrections(df=sample)

        for col in slow_pi_col_names:
            assert col not in sample.columns, f"{col}"
        assert len(high_pt_corr) == len(sample.index), f"{len(high_pt_corr)} and {len(sample.index)}"
        assert len(slow_pi_corr) == len(sample.index), f"{len(slow_pi_corr)} and {len(sample.index)}"

        sample.loc[:, self.tracking_systematic_name] = high_pt_corr
        for col in slow_pi_col_names:
            sample.loc[:, col] = slow_pi_corr[col].values

    @staticmethod
    def get_dstst_pi_var(variable: str) -> str:
        return f"daughter__bo1__cm__spdaughter__bo2__cm__sp{variable}__bc__bc"

    @staticmethod
    def get_track_infos() -> Dict[str, Dict[str, str]]:
        track_infos = {
            "lepton": {
                "pdg_column": ParticleIDCorrection.get_lepton_var("PDG"),
                "momentum_column": ParticleIDCorrection.get_lepton_var("p"),
                "theta_column": ParticleIDCorrection.get_lepton_var("theta"),
            }
        }

        track_infos.update(
            {
                "DststPi": {
                    "pdg_column": TrackingEffCorrector.get_dstst_pi_var("PDG"),
                    "momentum_column": TrackingEffCorrector.get_dstst_pi_var("p"),
                    "theta_column": TrackingEffCorrector.get_dstst_pi_var("theta"),
                }
            }
        )

        for index in range(0, 5):
            track_infos.update(
                {
                    f"D{index}": {
                        "pdg_column": ParticleIDCorrection.get_d_daughter_var(daughter=index, variable="PDG"),
                        "momentum_column": ParticleIDCorrection.get_d_daughter_var(daughter=index, variable="p"),
                        "theta_column": ParticleIDCorrection.get_d_daughter_var(daughter=index, variable="theta"),
                    }
                }
            )

        for index in range(0, 5):
            track_infos.update(
                {
                    f"Dst{index}": {
                        "pdg_column": ParticleIDCorrection.get_dd_daughter_var(daughter=index, variable="PDG"),
                        "momentum_column": ParticleIDCorrection.get_dd_daughter_var(daughter=index, variable="p"),
                        "theta_column": ParticleIDCorrection.get_dd_daughter_var(daughter=index, variable="theta"),
                    }
                }
            )

        return track_infos
