"""
Main Task for the D** analysis.

Patrick Ecker 2020
"""

import os
import b2luigi as luigi

from typing import Generator
from luigi.util import inherits

from rdstar.utilities import PathType
from rdstar.utilities.logging import set_logger, reset_third_party_logging_level

from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.tasks.migration_task import MigrationTask
from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection
from rdstar.offline_analysis.dss_fitting.tasks.bf_calculation_task import BFCalculationTask

# from rdstar.offline_analysis.dss_fitting.tasks.mva_optimizer_task import MVAOptimizationTask
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.dss_fitting.tasks.bf_toy_evaluation_task import BFToyEvaluationTask
from rdstar.offline_analysis.dss_fitting.tasks.ff_ratio_plotter_task import FFRatioCalculationTask
from rdstar.offline_analysis.dss_fitting.tasks.significance_plotter import SignificancePlotterTask


__all__ = [
    "FittingMaster",
]


logger = set_logger(name="rdstar_dss_offline_tasks", overwrite=True)
reset_third_party_logging_level()


@inherits(FittingPreparationTask)
class FittingMaster(luigi.WrapperTask):
    # dss_fit_selection = "dss_selection_dss_sec_sel_mva"  # type: str
    # dss_fit_selection = "dss_selection_full_pl_dss_sec_sel_mva"  # type: str
    # dss_fit_selection = "dss_selection_loose_mva_dss_sec_sel_mva"  # type: str
    dss_fit_selection = "dss_selection_full_pl_loose_mva_dss_sec_sel_mva"  # type: str

    dss_sub_dir_name = "DssFitting"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for fitting_setup in FittingSetupCollection.available_setups.keys():
            yield self.clone(FFRatioCalculationTask, fitting_setup_name=fitting_setup)
            yield self.clone(SignificancePlotterTask, fitting_setup_name=fitting_setup)
            if FittingSetupCollection.available_setups[fitting_setup].fit_in_bins_of.df_label == "qSquaredReco":
                yield self.clone(MigrationTask, fitting_setup_name=fitting_setup)

        for fitting_setup in BFFittingSetupCollection.available_setups.keys():
            yield self.clone(
                BFCalculationTask,
                fitting_setup_name=fitting_setup,
                plot_output=True,
                mva_cut_info=None,
            )
        for toy_fit_case_name in BFFitCaseDict.get_toy_fit_case_names():
            yield self.clone(
                BFToyEvaluationTask,
                fit_case_name=toy_fit_case_name,
                plot_output=True,
                mva_cut_info=None,
            )
            # yield self.clone(
            #     MVAOptimizationTask,
            #     fitting_setup_name=fitting_setup,
            # )


if __name__ == "__main__":

    os.environ["NUMEXPR_MAX_THREADS"] = "40"

    patrick_base_path = "/ceph/pecker/rdstar/fitting/"  # type: PathType
    production_key = "data_set_prod_02ndMay2021"  # type: str
    dss_selection = FittingMaster.dss_fit_selection  # type: str
    setup_appendix = "current"  # type: str

    work_dir = os.path.join(patrick_base_path, production_key, dss_selection, setup_appendix)  # type: PathType

    num_workers = 8
    luigi.process(
        FittingMaster(
            selection=FittingMaster.dss_fit_selection,
            production_key=production_key,
            working_dir=work_dir,
        ),
        workers=num_workers,
    )
