"""
Fit setup info for Continuum calibration fit.

Felix Metzner 2023
"""

from typing import Tuple
from uncertainties import ufloat

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    cont_cal_fit_obs_setup,
    base_cont_cal_weight_info,
    cont_cal_other_bbbar_bkg_sig_ids,
    cont_cal_dss_bkg_sig_ids,
    common_cont_cal_fit_component_id_column_name,
    add_common_cont_cal_fit_component_id_to,
    get_sig_ids_with_charge_info,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)


__all__ = [
    "cont_cal_fit_setup",
]


logger = log_to_default(name=__name__)


# Definition of components:

# (Main) Continuum Components

component_charged_d_reco_continuum_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="c_d",
    charge_type=None,
)
component_continuum_charged_d: ComponentInfo = ComponentInfo(
    name="continuum_in_charged_d",
    comp_index=10,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=component_charged_d_reco_continuum_ids,
)


component_neutral_d_reco_continuum_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="n_d",
    charge_type=None,
)
component_continuum_neutral_d: ComponentInfo = ComponentInfo(
    name="continuum_in_neutral_d",
    comp_index=11,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=component_neutral_d_reco_continuum_ids,
)


component_charged_ds_reco_continuum_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="c_dst",
    charge_type=None,
)
component_continuum_charged_ds: ComponentInfo = ComponentInfo(
    name="continuum_in_charged_ds",
    comp_index=12,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=component_charged_ds_reco_continuum_ids,
)


component_neutral_ds_reco_continuum_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="n_dst",
    charge_type=None,
)
component_continuum_neutral_ds: ComponentInfo = ComponentInfo(
    name="continuum_in_neutral_d",
    comp_index=13,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=component_neutral_ds_reco_continuum_ids,
)


# Background Components

# BKG in Charged D

cont_cal_Bp_Dz_l_wo_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="c_d",
    charge_type=None,
    cross_feed_type="wo_cf",
)

cont_cal_Bp_DzS_l_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="c_d",
    charge_type=None,
    cross_feed_type="cf",
)

cont_cal_charged_d_dss_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_dss_bkg_sig_ids,
    reco_type="c_d",
    charge_type=None,
)

cont_cal_charged_d_bkg_ids: Tuple[int, ...] = (
    *cont_cal_Bp_Dz_l_wo_cf_ids,
    *cont_cal_Bp_DzS_l_cf_ids,
    *cont_cal_charged_d_dss_bkg_ids,
)

component_charged_d_bkg: ComponentInfo = ComponentInfo(
    name="charged_d_BKG",
    comp_index=20,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_charged_d_bkg_ids,
)

cont_cal_charged_d_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_other_bbbar_bkg_sig_ids,
    reco_type="c_d",
    charge_type=None,
)
component_charged_d_other_bkg: ComponentInfo = ComponentInfo(
    name="charged_d_other_BKG",
    comp_index=30,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_charged_d_bbbar_bkg_ids,
)


# BKG in Neutral D

cont_cal_Bn_Dm_l_wo_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="n_d",
    charge_type=None,
    cross_feed_type="wo_cf",
)

cont_cal_Bn_DmS_l_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="n_d",
    charge_type=None,
    cross_feed_type="cf",
)

cont_cal_neutral_d_dss_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_dss_bkg_sig_ids,
    reco_type="n_d",
    charge_type=None,
)

cont_cal_neutral_d_bkg_ids: Tuple[int, ...] = (
    *cont_cal_Bn_Dm_l_wo_cf_ids,
    *cont_cal_Bn_DmS_l_cf_ids,
    *cont_cal_neutral_d_dss_bkg_ids,
)

component_neutral_d_bkg: ComponentInfo = ComponentInfo(
    name="neutral_d_BKG",
    comp_index=21,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_neutral_d_bkg_ids,
)

cont_cal_neutral_d_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_other_bbbar_bkg_sig_ids,
    reco_type="n_d",
    charge_type=None,
)
component_neutral_d_other_bkg: ComponentInfo = ComponentInfo(
    name="neutral_d_other_BKG",
    comp_index=31,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_neutral_d_bbbar_bkg_ids,
)


# BKG in Charged Ds

cont_cal_Bp_Dz_l_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="c_dst",
    charge_type=None,
    cross_feed_type="cf",
)

cont_cal_Bp_DzS_l_wo_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="c_dst",
    charge_type=None,
    cross_feed_type="wo_cf",
)

cont_cal_charged_ds_dss_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_dss_bkg_sig_ids,
    reco_type="c_dst",
    charge_type=None,
)

cont_cal_charged_ds_bkg_ids: Tuple[int, ...] = (
    *cont_cal_Bp_Dz_l_cf_ids,
    *cont_cal_Bp_DzS_l_wo_cf_ids,
    *cont_cal_charged_ds_dss_bkg_ids,
)

component_charged_ds_bkg: ComponentInfo = ComponentInfo(
    name="charged_ds_BKG",
    comp_index=22,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_charged_ds_bkg_ids,
)

cont_cal_charged_ds_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_other_bbbar_bkg_sig_ids,
    reco_type="c_dst",
    charge_type=None,
)
component_charged_ds_other_bkg: ComponentInfo = ComponentInfo(
    name="charged_ds_other_BKG",
    comp_index=32,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_charged_ds_bbbar_bkg_ids,
)


# BKG in Neutral Ds

cont_cal_Bz_Dm_l_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="n_dst",
    charge_type=None,
    cross_feed_type="cf",
)

cont_cal_Bz_DmS_l_wo_cf_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="n_dst",
    charge_type=None,
    cross_feed_type="wo_cf",
)

cont_cal_neutral_ds_dss_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_dss_bkg_sig_ids,
    reco_type="n_dst",
    charge_type=None,
)

cont_cal_neutral_ds_bkg_ids: Tuple[int, ...] = (
    *cont_cal_Bz_Dm_l_cf_ids,
    *cont_cal_Bz_DmS_l_wo_cf_ids,
    *cont_cal_neutral_ds_dss_bkg_ids,
)

component_neutral_ds_bkg: ComponentInfo = ComponentInfo(
    name="neutral_ds_BKG",
    comp_index=23,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_neutral_ds_bkg_ids,
)

cont_cal_neutral_ds_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=cont_cal_other_bbbar_bkg_sig_ids,
    reco_type="n_dst",
    charge_type=None,
)
component_neutral_ds_other_bkg: ComponentInfo = ComponentInfo(
    name="neutral_ds_other_BKG",
    comp_index=33,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=cont_cal_neutral_ds_bbbar_bkg_ids,
)


cont_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    # (Main) Continuum Components
    component_continuum_charged_d,
    component_continuum_neutral_d,
    component_continuum_charged_ds,
    component_continuum_neutral_ds,
    # (Bkg) Normalizaion and D** Components
    component_charged_d_bkg,
    component_neutral_d_bkg,
    component_charged_ds_bkg,
    component_neutral_ds_bkg,
    # (Bkg) Other BBbar Components
    component_charged_d_other_bkg,
    component_neutral_d_other_bkg,
    component_charged_ds_other_bkg,
    component_neutral_ds_other_bkg,
)


cont_cal_fit_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="Bz_to_Dm",
            ch_index=0,
            reco_mode_info=RecoModeInfos.neutral_b_to_d_modes,
            components=cont_cal_fit_comps,
            observable_infos=cont_cal_fit_obs_setup.cont_obs_info_Bz_to_Dm,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bm_to_Dz",
            ch_index=1,
            reco_mode_info=RecoModeInfos.charged_b_to_d_modes,
            components=cont_cal_fit_comps,
            observable_infos=cont_cal_fit_obs_setup.cont_obs_info_Bm_to_Dz,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bz_to_Dsm",
            ch_index=2,
            reco_mode_info=RecoModeInfos.neutral_b_to_ds_modes,
            components=cont_cal_fit_comps,
            observable_infos=cont_cal_fit_obs_setup.cont_obs_info_Bz_to_Dsm,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bm_to_Dsz",
            ch_index=3,
            reco_mode_info=RecoModeInfos.charged_b_to_ds_modes,
            components=cont_cal_fit_comps,
            observable_infos=cont_cal_fit_obs_setup.cont_obs_info_Bm_to_Dsz,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
    ]
)


cont_cal_fit_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[],
    key_order=(),
)


_cont_cal_asimov_fit_setup = FitSetupInfoContainer(
    components=cont_cal_fit_comps,
    reco_channels=cont_cal_fit_chs,
    observable_infos=cont_cal_fit_obs_setup.cont_cal_obs_infos,
    weight_info=base_cont_cal_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=cont_cal_fit_tag_eff_corr_info,
    reco_mode_column_name=Cols.decay_mode_column_name,
    component_id_column_name=common_cont_cal_fit_component_id_column_name,
    component_id_function=add_common_cont_cal_fit_component_id_to,
)


cont_cal_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_cont_cal_asimov_fit_setup)
