"""
Fit observable infos for fit setups of dedicated R(D) and R(D*) fit.

Felix Metzner 2022
"""

import numpy as np

from typing import Tuple, Dict

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitObservableInfo


__all__ = [
    "binning_info_for_reco_modes_dict",
    "binning_info_for_reco_modes_obs_fit_mms_dict",
    "binning_info_for_reco_modes_obs_fit_neextra_dict",
    #
    "FitObsSetupMain",
    "FitObsSetupMainWithFineBinning",
    "FitObsSetup1",
    "FitObsSetup1p1",
    "FitObsSetup1p2",
    "FitObsSetup2",
    "FitObsSetup2p1",
    "FitObsSetup2p2",
    "FitObsSetup2p3",
    "FitObsSetup3",
    "FitObsSetup3p1",
    "FitObsSetup3p2",
    "FitObsSetup3p3",
    "FitObsSetup3p4",
    "FitObsSetup4",
    "FitObsSetup4p1",
    "FitObsSetup4p2",
    "FitObsSetup4p3",
    "FitObsSetup4p4",
    "FitObsSetup4p5",
    "FitObsSetup4p6",
    "FitObsSetup4p7",
    "FitObsSetup4p8",
    "FitObsSetup4p9",
    "FitObsSetup4p10",
    "FitObsSetup4p11",
    "FitObsSetup4p12",
    "FitObsSetup4p13",
    "FitObsSetup5",
    "FitObsSetup6",
    "FitObsSetup6p1",
    "FitObsSetup6p2",
]


logger = log_to_default(name=__name__)


original_obs_fit_mms_binning: np.ndarray = np.linspace(
    *RDStarHistVars.missing_mass_res_corr_for_main_fit.scope,
    RDStarHistVars.missing_mass_res_corr_for_main_fit.n_bins + 1,
)
original_obs_fit_neextra_binning: np.ndarray = np.linspace(
    *RDStarHistVars.neextra_for_main_fit.scope,
    RDStarHistVars.neextra_for_main_fit.n_bins + 1,
)
original_binning_tuple: Tuple[np.ndarray, np.ndarray] = (original_obs_fit_mms_binning, original_obs_fit_neextra_binning)

obs_fit_mms_nB_to_D_binning: np.ndarray = np.array([-0.5, 0.5, 1.5, 2.5, 3.5, 5.5, 9.5])
obs_fit_neextra_nB_to_D_binning: np.ndarray = np.array([0.0, 0.14, 0.28, 0.42, 0.56, 0.84, 1.12, 1.4])
binning_info_nB_to_D: Tuple[np.ndarray, np.ndarray] = (obs_fit_mms_nB_to_D_binning, obs_fit_neextra_nB_to_D_binning)

obs_fit_mms_cB_to_D_binning: np.ndarray = np.array([-0.5, 0.5, 1.5, 2.5, 3.5, 5.5, 9.5])
obs_fit_neextra_cB_to_D_binning: np.ndarray = np.array([0.0, 0.14, 0.28, 0.42, 0.56, 0.70, 0.84, 0.98, 1.12, 1.26, 1.4])
binning_info_cB_to_D: Tuple[np.ndarray, np.ndarray] = (obs_fit_mms_cB_to_D_binning, obs_fit_neextra_cB_to_D_binning)

obs_fit_mms_nB_to_Ds_binning: np.ndarray = np.array([-0.5, 0.5, 1.5, 9.5])
obs_fit_neextra_nB_to_Ds_binning: np.ndarray = np.array([0.0, 0.40, 0.80, 1.4])
binning_info_nB_to_Ds: Tuple[np.ndarray, np.ndarray] = (obs_fit_mms_nB_to_Ds_binning, obs_fit_neextra_nB_to_Ds_binning)

obs_fit_mms_cB_to_Ds_binning: np.ndarray = np.array([-0.5, 0.5, 1.5, 2.5, 3.5, 9.5])
# obs_fit_neextra_cB_to_Ds_binning: np.ndarray = np.array([0.0, 0.14, 0.28, 0.42, 0.56, 0.70, 0.84, 1.4])
obs_fit_neextra_cB_to_Ds_binning: np.ndarray = np.array([0.0, 0.14, 0.28, 0.42, 0.56, 0.84, 1.12, 1.4])
binning_info_cBs_to_D: Tuple[np.ndarray, np.ndarray] = (obs_fit_mms_cB_to_Ds_binning, obs_fit_neextra_cB_to_Ds_binning)


binning_info_for_reco_modes_dict: Dict[str, Tuple[np.ndarray, np.ndarray]] = {
    RecoModeInfos.neutral_b_to_d_modes.name: binning_info_nB_to_D,
    RecoModeInfos.charged_b_to_d_modes.name: binning_info_cB_to_D,
    RecoModeInfos.neutral_b_to_ds_modes.name: binning_info_nB_to_Ds,
    RecoModeInfos.charged_b_to_ds_modes.name: binning_info_cBs_to_D,
    RecoModeInfos.get_all_reco_modes_latex_label_key(): original_binning_tuple,
}


binning_info_for_reco_modes_obs_fit_mms_dict: Dict[str, Tuple[float, ...]] = {
    k: tuple(v[0]) for k, v in binning_info_for_reco_modes_dict.items()
}
binning_info_for_reco_modes_obs_fit_neextra_dict: Dict[str, Tuple[float, ...]] = {
    k: tuple(v[1]) for k, v in binning_info_for_reco_modes_dict.items()
}


class FitObsSetupMain:
    binning_for_reco_modes: Dict[str, Tuple[np.ndarray, np.ndarray]] = binning_info_for_reco_modes_dict
    binning_for_reco_modes_obs_fit_mms: Dict[str, Tuple[float, ...]] = binning_info_for_reco_modes_obs_fit_mms_dict
    binning_for_reco_modes_obs_fit_neextra: Dict[
        str, Tuple[float, ...]
    ] = binning_info_for_reco_modes_obs_fit_neextra_dict

    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_main_fit,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr_for_main_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=tuple(obs_fit_neextra_nB_to_D_binning)),
        obs2.reset(scope=None, binning=tuple(obs_fit_mms_nB_to_D_binning)),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=tuple(obs_fit_neextra_cB_to_D_binning)),
        obs2.reset(scope=None, binning=tuple(obs_fit_mms_cB_to_D_binning)),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=tuple(obs_fit_neextra_nB_to_Ds_binning)),
        obs2.reset(scope=None, binning=tuple(obs_fit_mms_nB_to_Ds_binning)),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=tuple(obs_fit_neextra_cB_to_Ds_binning)),
        obs2.reset(scope=None, binning=tuple(obs_fit_mms_cB_to_Ds_binning)),
    )


class FitObsSetupMainWithFineBinning:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_main_fit,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr_for_main_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=10),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=9),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup1p1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 8)),
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=8),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=12),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=6),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=10),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup1p2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 16)),
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=28),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=16),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=20),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=10),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=9),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup2p1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=(0.808, 1.0), binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=12),
        obs2.reset(scope=(0.808, 1.0), binning=10),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=9),
        obs2.reset(scope=(0.575, 1.0), binning=14),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=(0.575, 1.0), binning=14),
    )


class FitObsSetup2p2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=12),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=9),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=11),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )


class FitObsSetup2p3:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=24),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=24),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=24),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=24),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )


class FitObsSetup3:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup3p1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.808, 1.0), binning=10),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.808, 1.0), binning=10),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.575, 1.0), binning=14),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.575, 1.0), binning=14),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup3p2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup3p3:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 20)),
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup3p4:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 20))),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=24),
        obs2.reset(scope=None, binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 24))),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=16),
        obs2.reset(scope=None, binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 16))),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=(0.2, 1.0), binning=20),
        obs2.reset(scope=None, binning=tuple(np.linspace(*RDStarHistVars.pl_star_signal_fit.scope, 20))),
    )


class FitObsSetup4:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup4p1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=9),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )


class FitObsSetup4p2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )


class FitObsSetup4p3:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=None,
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=9),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )


class FitObsSetup4p4:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(0.0, 2.5, 26)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=9),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )


class FitObsSetup4p5:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.2, 1.6, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )


class FitObsSetup4p6:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 10)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=8),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=6),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )


class FitObsSetup4p7:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(sorted(list(np.linspace(0.0, 2.5, 11)) + [0.1, 0.2])),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=9),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )


class FitObsSetup4p8:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 10)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=9),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=11),
    )


class FitObsSetup4p9:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 18)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=28),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=20),
    )


class FitObsSetup4p10:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 18)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=28),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=20),
    )


class FitObsSetup4p11:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 18)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=16),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=18),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=22),
    )


class FitObsSetup4p12:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 18)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )


class FitObsSetup4p13:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 15)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.missing_mass_res_corr,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=24),
    )


class FitObsSetup5:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.pl_star_signal_fit,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=None, binning=None),
    )


class FitObsSetup6:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.808, 1.0), binning=10),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.808, 1.0), binning=10),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.575, 1.0), binning=14),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.575, 1.0), binning=14),
    )


class FitObsSetup6p1:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=(0.0, 0.04, 0.8, 0.12, 0.16, 0.2, 0.3, 0.4, 0.6, 1.0, 1.5, 2.0, 2.5),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )


class FitObsSetup6p2:
    obs1 = FitObservableInfo(
        hist_var=RDStarHistVars.neextra_for_fit,
        alt_scope=None,
        alt_binning=tuple(np.linspace(*RDStarHistVars.neextra_for_fit.scope, 18)),
    )
    obs2 = FitObservableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_fit_var,
        alt_scope=None,
        alt_binning=None,
    )

    rdstar_obs_infos: Tuple[FitObservableInfo, ...] = (
        obs1,
        obs2,
    )

    rdstar_obs_info_Bz_to_Dm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bz_to_Dsm: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
    rdstar_obs_info_Bm_to_Dsz: Tuple[FitObservableInfo, ...] = (
        obs1.reset(scope=None, binning=None),
        obs2.reset(scope=(0.2, 1.0), binning=20),
    )
