"""
Base fitter class for dedicated fit approach.

Felix Metzner 2022
"""

import copy
import numpy as np
import pandas as pd

from iminuit import Minuit
from dataclasses import dataclass
from collections import defaultdict
from abc import ABC, abstractmethod
from scipy import stats as scipy_stats
from uncertainties import UFloat, ufloat
from typing import List, Tuple, Dict, Optional, Iterator, MutableMapping, Any

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import SimpleEfficiencyCalculator
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    RecoChannelInfo,
)


__all__ = [
    "FitterBase",
    "LocalEfficiencyInfo",
    "LocalEfficiencyInfoContainer",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class MinuitSettings:
    error_factor: float
    error_def: float
    print_level: int
    n_call: int
    n_iterate: int

    def __post_init__(self):
        assert self.error_def in [Minuit.LIKELIHOOD, Minuit.LEAST_SQUARES], (
            self.error_def,
            [Minuit.LIKELIHOOD, Minuit.LEAST_SQUARES],
        )

        assert self.n_iterate > 0, self.n_iterate
        assert self.n_call > 0, self.n_call
        assert self.error_factor > 0, self.error_factor


default_minuit_settings = MinuitSettings(
    error_factor=0.05,
    error_def=Minuit.LEAST_SQUARES,
    print_level=1,
    n_call=100_000,
    n_iterate=2,
)


@dataclass(frozen=True)
class LocalEfficiencyInfo:
    component_id: int
    reco_channel_id: int
    n_before: UFloat
    n_after: UFloat
    n_before_selection: UFloat

    @property
    def efficiency(self) -> float:
        if self.n_before.nominal_value:
            eff: UFloat = self.n_after / self.n_before
            return eff.nominal_value
        else:
            return 0.0


class LocalEfficiencyInfoContainer(MutableMapping[Tuple[int, int], LocalEfficiencyInfo]):
    def __init__(self) -> None:
        self._local_effs: Dict[Tuple[int, int], LocalEfficiencyInfo] = dict()
        self._components_map: Dict[int, List[Tuple[int, int]]] = defaultdict(list)
        self._reco_chs_map: Dict[int, List[Tuple[int, int]]] = defaultdict(list)

    def __getitem__(self, item: Tuple[int, int]) -> LocalEfficiencyInfo:
        return self._local_effs[item]

    def __setitem__(self, key: Tuple[int, int], value: LocalEfficiencyInfo) -> None:
        raise RuntimeError("LocalEfficiencyInfoContainer is append-only.")

    def __delitem__(self, key: Tuple[int, int]) -> None:
        raise RuntimeError("LocalEfficiencyInfoContainer is append-only.")

    def __iter__(self) -> Iterator[Tuple[int, int]]:
        return iter(self._local_effs)

    def __len__(self) -> int:
        return len(self._local_effs)

    def append(self, item: LocalEfficiencyInfo) -> None:
        key: Tuple[int, int] = (item.component_id, item.reco_channel_id)
        assert key not in self._local_effs, (key, self._local_effs.keys())
        self._local_effs[key] = item
        self._components_map[item.component_id].append(key)
        self._reco_chs_map[item.reco_channel_id].append(key)

    def check_consistency(self) -> None:
        for c_id, keys in self._components_map.items():
            assert len(keys) == len(set(keys)), keys

            assert len(set([self[k].n_before.n for k in keys])) == 1, (c_id, [self[k].n_before for k in keys])

            _n_before: float = self[keys[0]].n_before.nominal_value
            _list_n_after: List[float] = [self[k].n_after.nominal_value for k in keys]
            assert np.isclose(sum(_list_n_after), _n_before), (c_id, _n_before, _list_n_after, sum(_list_n_after))

            eff_values = [self[k].efficiency for k in keys]
            assert np.isclose(sum(eff_values), 1.0), (sum(eff_values), eff_values, c_id, keys)

    def get_initial_sum_of_weights_per_component(self) -> Tuple[Tuple[int, float], ...]:
        _res: List[Tuple[int, float]] = [
            (c_id, self[info_tuples[0]].n_before.nominal_value) for c_id, info_tuples in self._components_map.items()
        ]
        return tuple(_res)

    def get_error_on_sum_of_weights_per_component(self) -> Tuple[Tuple[int, float], ...]:
        _res: List[Tuple[int, float]] = [
            (c_id, self[info_tuples[0]].n_before.std_dev) for c_id, info_tuples in self._components_map.items()
        ]
        return tuple(_res)


class FitterBase(ABC):

    ordered_parameter_names: Tuple[str, ...] = tuple([])

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        file_dict_dir_path: Optional[PathType] = None,
        **kwargs,
    ) -> None:
        if ignore:
            raise TypeError("Only keyword arguments allowed!")
        if kwargs:
            raise TypeError(f"Unknown keyword argument(s) {[k for k in kwargs]}!")

        self.minuit_settings: MinuitSettings = default_minuit_settings

        self.file_dict_dir_path: Optional[PathType] = file_dict_dir_path

        self._fit_setup_info: FitSetupInfoContainer = fit_setup_info
        self._mc_normalization_factor: float = 1.0

        self._template_shapes: Optional[np.ndarray] = None
        self._base_mc_bin_counts: Optional[np.ndarray] = None
        self._base_mc_bin_counts_for_uncerts: Optional[np.ndarray] = None
        self._pure_mc_templates: Optional[np.ndarray] = None
        self._squared_template_stat_error: Optional[np.ndarray] = None

        self._bin_indices_range_per_reco_channel: Optional[Dict[int, Tuple[int, int]]] = None

        self._empty_bin_mask: Optional[np.ndarray] = None
        self._final_n_bins_total: Optional[int] = None

        self._bin_stat_errors_squared: Optional[np.ndarray] = None

        self._local_efficiency_infos: LocalEfficiencyInfoContainer = LocalEfficiencyInfoContainer()
        self._local_efficiencies: Optional[np.ndarray] = None

        self._random_state_for_toy_seeds: Optional[np.random.RandomState] = None
        self._base_toy_distribution: Optional[np.ndarray] = None

        self._data_weight_col_name: Optional[str] = None  # TODO: This is nowhere defined!
        self._data_df_weights: Optional[np.ndarray] = None

        self._data_bin_counts: Optional[np.ndarray] = None
        self._data_bin_stat_errors_squared: Optional[np.ndarray] = None

        self._mc_df: pd.DataFrame = self._init_mc_data_frame(mc_df=mc_df)
        self._data_df: Optional[pd.DataFrame] = self._init_data_data_frame(data_df=data_df)

        self._params_vector: np.ndarray = np.array([-999])
        self._initial_params_vector: np.ndarray = np.array([-999])

        self._global_eff_info: Optional[SimpleEfficiencyCalculator] = None

    # region General info
    @property
    def fit_setup_info(self) -> FitSetupInfoContainer:
        return self._fit_setup_info

    @property
    def internal_mc_sample_df(self) -> pd.DataFrame:
        return self._mc_df

    # endregion

    # region DataFrame initialization
    # MC data frame
    def _init_mc_data_frame(self, mc_df: pd.DataFrame) -> pd.DataFrame:
        logger.info("Initializing mc_df")
        mask: np.ndarray = np.full(shape=len(mc_df), fill_value=False)

        _reco_ch_infos_for_local_effs: List[Tuple[RecoChannelInfo, np.ndarray]] = list()

        for reco_ch_info in self.fit_setup_info.reco_channels:
            logger.info(f"In initialization of mc_df: getting info for reco_ch {reco_ch_info.name} mask")
            reco_ch_mask: np.ndarray = reco_ch_info.get_mask(
                df=mc_df,
                reco_mode_column_str=self.fit_setup_info.reco_mode_col,
            )

            _reco_ch_infos_for_local_effs.append((reco_ch_info, reco_ch_mask))

            mask |= reco_ch_mask

        self.calculate_local_efficiency_infos(
            df=mc_df,
            reco_ch_infos=_reco_ch_infos_for_local_effs,
            combined_mask=mask,
        )

        return mc_df[mask].copy(deep=True)

    # Measured data points
    def _init_data_data_frame(
        self,
        data_df: Optional[pd.DataFrame],
    ) -> Optional[pd.DataFrame]:

        if self.fit_setup_info.data_setup_info.recorded_data:
            logger.info("Initializing data_df from data...")
            return self._init_data_df_from_recorded_data(data_df=data_df)
        else:
            logger.info("Initializing data_df from mc")
            assert data_df is None
            self._init_data_df_from_mc()
            return None

    # Real data points from recorded data frame
    def _init_data_df_from_recorded_data(self, data_df: Optional[pd.DataFrame]) -> pd.DataFrame:
        assert data_df is not None

        if self.fit_setup_info.reco_mode_col == TagRecoModeInfos.fei_reco_mode_col_name:
            if self.fit_setup_info.reco_mode_col not in data_df.columns:
                TagRecoModeInfos.add_new_fei_reco_mode_col(df=data_df, check_if_new=True)

        if self.fit_setup_info.reco_mode_col == TagRecoModeInfos.fei_plus_sig_d_reco_mode_col_name:
            if self.fit_setup_info.reco_mode_col not in data_df.columns:
                TagRecoModeInfos.add_sig_d_plus_fei_reco_mode_col(df=data_df, check_if_new=True)

        mask: np.ndarray = np.full(shape=len(data_df), fill_value=False)
        for reco_ch_info in self.fit_setup_info.reco_channels:
            mask |= reco_ch_info.get_mask(df=data_df, reco_mode_column_str=self.fit_setup_info.reco_mode_col)

        return data_df[mask]

    # Simulated data points
    def _init_data_df_from_mc(self) -> None:
        assert self._data_bin_counts is None
        assert self._data_bin_stat_errors_squared is None

        if self.fit_setup_info.data_setup_info.asimov_data:
            logger.info("Initializing data_df from mc... as asimov data")
            _bin_count: np.ndarray = self._get_data_df_from_asimov_data()
        elif self.fit_setup_info.data_setup_info.toy_data:
            logger.info("Initializing data_df from mc... as toy data")
            assert self._random_state_for_toy_seeds is None
            self._random_state_for_toy_seeds = np.random.RandomState(seed=self.fit_setup_info.data_setup_info.toy_seed)
            _bin_count = self._get_data_df_from_toy_data()
        else:
            raise RuntimeError("Unknown data setup type!")

        self._check_shape_dim_zero(a=_bin_count, after_zero_bin_mask=True, before_ch_concat=False)
        self._check_shape_dim_one(a=_bin_count, is_single_dim=True)
        self._check_shape_dim_two(a=_bin_count, has_less_dim=True)

        self._data_bin_counts = _bin_count
        self._data_bin_stat_errors_squared = copy.deepcopy(_bin_count)

    # Asimov data points
    def _get_data_df_from_asimov_data(self) -> np.ndarray:
        _asimov_data_bin_counts: np.ndarray = np.sum(self.pure_mc_templates, axis=1)

        if self.fit_setup_info.data_setup_info.round_bin_count:
            _asimov_data_bin_counts = np.ceil(_asimov_data_bin_counts)

        return _asimov_data_bin_counts

    # Toy data points
    @property
    def next_toy_seed(self) -> int:
        assert self._random_state_for_toy_seeds is not None
        assert self.fit_setup_info.data_setup_info.toy_size is not None
        return scipy_stats.randint.rvs(
            low=0,
            high=self.fit_setup_info.data_setup_info.toy_size + 100_000,
            loc=0,
            size=self.fit_setup_info.data_setup_info.toy_size,
            random_state=self._random_state_for_toy_seeds,
        )

    def _get_next_toy_sample(self) -> np.ndarray:
        if self._base_toy_distribution is None:
            self._base_toy_distribution = np.sum(self.pure_mc_templates, axis=1)
        _random_state: np.random.RandomState = np.random.RandomState(seed=self.next_toy_seed)
        return scipy_stats.poisson.rvs(self._base_toy_distribution, random_state=_random_state)

    def _get_data_df_from_toy_data(self) -> np.ndarray:
        if self.fit_setup_info.data_setup_info.round_bin_count:
            return np.ceil(self._get_next_toy_sample())
        else:
            return self._get_next_toy_sample()

    def set_next_toy_sample(self) -> None:
        if self.fit_setup_info.data_setup_info.round_bin_count:
            _toy_bin_count: np.ndarray = np.ceil(self._get_next_toy_sample())
        else:
            _toy_bin_count = self._get_next_toy_sample()
        self._data_bin_counts = _toy_bin_count
        self._data_bin_stat_errors_squared = copy.deepcopy(_toy_bin_count)

    # endregion

    # region Components
    @property
    def number_of_components(self) -> int:
        return self.fit_setup_info.n_components

    # endregion

    # region Binning, variables and weights
    # Variable methods
    @property
    def obs_column_names(self) -> Tuple[str, ...]:
        return tuple(obs.col_name for obs in self.fit_setup_info.observable_infos)

    # Binning methods
    @property
    def final_n_bins_total(self) -> int:
        if self._final_n_bins_total is None:
            assert self.empty_bin_mask is not None
            self._final_n_bins_total = self.fit_setup_info.max_n_bins_total - np.sum(self.empty_bin_mask)
        return self._final_n_bins_total

    # Weight Methods
    @property
    def mc_weight_col_name(self) -> str:
        return self.fit_setup_info.weight_info.col_name

    @property
    def mc_normalization_factor(self):
        return self._mc_normalization_factor

    # endregion

    # region Templates
    def calculate_base_template_shapes(self) -> None:
        if self._template_shapes is not None:
            raise RuntimeError("Template shapes (2D) are already defined...")

        _template_shapes = self._get_reco_ch_concatenated_base_templates(
            templates_per_ch=self.base_mc_bin_counts,
            normalize=True,
        )

        assert not np.any(np.isnan(_template_shapes)), np.sum(np.isnan(_template_shapes))

        self._check_shape_dim_zero(a=_template_shapes, after_zero_bin_mask=False, before_ch_concat=False)
        self._check_shape_dim_one(a=_template_shapes, is_single_dim=False)
        self._check_shape_dim_two(a=_template_shapes, has_less_dim=True)

        if self.fit_setup_info.remove_empty_bins:
            _template_shapes = np.delete(_template_shapes, self.empty_bin_mask, axis=0)

        self._check_shape_dim_zero(a=_template_shapes, after_zero_bin_mask=True, before_ch_concat=False)

        self._template_shapes = _template_shapes

    def calculate_squared_mc_stat_errors(self, for_templates: bool = False) -> np.ndarray:
        _base_bin_stat_errors_squared: np.ndarray = self.base_mc_bin_counts_for_uncertainties

        self.check_base_mc_bin_counts(a=_base_bin_stat_errors_squared)

        _reco_ch_concatenated_stat_errors_squared: np.ndarray = self._get_reco_ch_concatenated_base_templates(
            templates_per_ch=_base_bin_stat_errors_squared,
            normalize=False,
        )

        if self.fit_setup_info.remove_empty_bins:
            _reco_ch_concatenated_stat_errors_squared = np.delete(
                _reco_ch_concatenated_stat_errors_squared,
                self.empty_bin_mask,
                axis=0,
            )

        self._check_shape_dim_one(a=_reco_ch_concatenated_stat_errors_squared, is_single_dim=False)
        self._check_shape_dim_zero(
            a=_reco_ch_concatenated_stat_errors_squared,
            after_zero_bin_mask=True,
            before_ch_concat=False,
        )

        if for_templates:
            return _reco_ch_concatenated_stat_errors_squared
        else:
            _bin_stat_errors_squared: np.ndarray = np.sum(_reco_ch_concatenated_stat_errors_squared, axis=1)

        return _bin_stat_errors_squared

    @property
    def mc_bin_stat_errors_squared(self) -> np.ndarray:
        if self._bin_stat_errors_squared is None:
            self._bin_stat_errors_squared = self.calculate_squared_mc_stat_errors(for_templates=False)
        return self._bin_stat_errors_squared

    def _check_shape_dim_zero(
        self,
        a: np.ndarray,
        after_zero_bin_mask: bool,
        before_ch_concat: bool,
    ) -> None:
        assert not (before_ch_concat and after_zero_bin_mask), a.shape
        if after_zero_bin_mask and not before_ch_concat:
            assert self.final_n_bins_total == a.shape[0], (
                self.final_n_bins_total,
                self.fit_setup_info.n_bins_total,
                np.sum(self.empty_bin_mask),
                a.shape,
            )
        elif before_ch_concat and not after_zero_bin_mask:
            assert self.fit_setup_info.max_n_flattened_bins_per_ch == a.shape[0], (
                self.fit_setup_info.max_n_flattened_bins_per_ch,
                a.shape,
            )
        else:
            assert self.fit_setup_info.max_n_bins_total == a.shape[0], (self.fit_setup_info.max_n_bins_total, a.shape)

    def _check_shape_dim_one(self, a: np.ndarray, is_single_dim: bool) -> None:
        if not is_single_dim:
            assert self.number_of_components == a.shape[1], (self.number_of_components, a.shape)
        else:
            assert len(a.shape) == 1, a.shape

    def _check_shape_dim_two(self, a: np.ndarray, has_less_dim: bool) -> None:
        if has_less_dim:
            assert len(a.shape) <= 2, a.shape
        else:
            assert self.fit_setup_info.n_reco_channels == a.shape[2], (self.fit_setup_info.n_reco_channels, a.shape)

    def _get_padding_for_channel(self, reco_ch_info: RecoChannelInfo) -> Tuple[int, int]:
        padding_width: int = self.fit_setup_info.max_n_flattened_bins_per_ch - reco_ch_info.n_bins_total
        assert padding_width >= 0, (
            padding_width,
            self.fit_setup_info.max_n_flattened_bins_per_ch,
            reco_ch_info.n_bins_total,
        )
        return 0, padding_width

    def get_bin_indices_range_per_reco_channel(self) -> Dict[int, Tuple[int, int]]:
        if self._bin_indices_range_per_reco_channel is None:
            _index_range_map: Dict[int, Tuple[int, int]] = {}

            n_reco_chs: int = len(self.fit_setup_info.reco_channels)
            expected_length_of_empty_bin_mask: int = self.fit_setup_info.max_n_flattened_bins_per_ch * n_reco_chs

            assert len(self.empty_bin_mask) == expected_length_of_empty_bin_mask, (
                len(self.empty_bin_mask),
                expected_length_of_empty_bin_mask,
            )

            _last_channels_last_id: int = 0

            for ch_i, reco_ch in enumerate(self.fit_setup_info.reco_channels):
                _max_bin_ids_start: int = ch_i * self.fit_setup_info.max_n_flattened_bins_per_ch
                _max_bin_ids_end: int = (ch_i + 1) * self.fit_setup_info.max_n_flattened_bins_per_ch
                _n_non_empty_bins_in_this_ch = np.sum(~self.empty_bin_mask[_max_bin_ids_start:_max_bin_ids_end])

                _this_channels_last_id: int = _last_channels_last_id + _n_non_empty_bins_in_this_ch

                ch_index_range: Tuple[int, int] = (_last_channels_last_id, _this_channels_last_id)

                _index_range_map.update({reco_ch.ch_index: ch_index_range})
                _last_channels_last_id = _this_channels_last_id

            assert _last_channels_last_id == self.template_shapes.shape[0], (
                _last_channels_last_id,
                self.template_shapes.shape,
                _index_range_map,
            )

            assert len(_index_range_map) == len(self.fit_setup_info.reco_channels), (
                len(_index_range_map),
                len(self.fit_setup_info.reco_channels),
                list(_index_range_map.keys()),
                [rch.ch_index for rch in self.fit_setup_info.reco_channels],
            )

            self._bin_indices_range_per_reco_channel = _index_range_map

        return self._bin_indices_range_per_reco_channel

    def _calculate_base_mc_bin_counts(
        self,
        for_uncertainties: bool,
    ) -> np.ndarray:
        """
        This method will generate the un-normalized base template numpy array with the shape
        (n_bins_per_channel_max, n_components, n_channels)
        where:
            - n_bins_per_channel_max is the maximum over all channels for the number of bins of the flattened templates,
            - n_components is the number of components, and
            - n_channels is the number of reconstruction channels.
        The latter is later factored out by flattening along the n_channel dimension, to get the shape
        (n_bins_per_channel_max * n_channels, n_components)
        To make this possible, the templates for a specific channel must be padded to n_bins_per_channel_max.
        The padding will later be removed by selecting only non-empty bins, yielding the final shape
        (n_non_empty_bins_in_all_channels, n_components)
        where
            n_non_empty_bins_in_all_channels = n_bins_per_channel_max * n_channels - n_empty_bins
        """
        logger.info(f"Generating base MC bin counts{' for uncertainties' if for_uncertainties else ''}")
        templates_per_ch: List[np.ndarray] = []

        for reco_ch in self.fit_setup_info.reco_channels:
            logger.info(
                f"In generation of base MC bin counts{' for uncertainties' if for_uncertainties else ''}: "
                f"looking at reco channel {reco_ch.name}"
            )

            reco_ch_mask: np.ndarray = reco_ch.get_mask(
                df=self._mc_df,
                reco_mode_column_str=self.fit_setup_info.reco_mode_col,
            )

            _channel_padding: Tuple[int, int] = self._get_padding_for_channel(reco_ch_info=reco_ch)

            templates_per_comp: List[np.ndarray] = []
            for component in self.fit_setup_info.components:
                comp_mask: np.ndarray = component.get_component_mask(
                    df=self._mc_df,
                    component_id_column_name=self.fit_setup_info.component_id_column_name,
                )
                _mask: np.ndarray = comp_mask & reco_ch_mask

                if component not in reco_ch.components or not np.any(_mask):
                    _template: np.ndarray = np.zeros(shape=reco_ch.n_bins_per_obs)

                else:
                    _weight_values: np.ndarray = self._mc_df[_mask][self.mc_weight_col_name].values
                    assert len(_weight_values.shape) == 1, _weight_values.shape

                    if for_uncertainties:
                        _weight_values = np.square(_weight_values * self.mc_normalization_factor)

                    _template = np.histogramdd(
                        sample=self._mc_df[_mask][list(self.obs_column_names)].values,
                        weights=_weight_values,
                        bins=reco_ch.n_bins_per_obs,
                        range=reco_ch.scope_per_obs,
                    )[0]

                templates_per_comp.append(
                    np.pad(_template.flatten(), pad_width=_channel_padding, mode="constant", constant_values=0.0)
                    if _channel_padding[1] > 0
                    else _template.flatten()
                )

            templates_per_ch.append(np.stack(templates_per_comp).T)

        logger.info(f"Finished generation of base MC bin counts{' for uncertainties' if for_uncertainties else ''}")
        return np.stack(templates_per_ch, axis=2)

    @staticmethod
    def _get_reco_ch_concatenated_base_templates(templates_per_ch: np.ndarray, normalize: bool) -> np.ndarray:
        if normalize:
            _norm: np.ndarray = np.sum(templates_per_ch, axis=0)[np.newaxis, :, :]

            if np.any(_norm == 0.0):
                _temps: np.ndarray = np.where(_norm > 0.0, templates_per_ch / _norm, templates_per_ch)
            else:
                _temps = templates_per_ch / _norm

            return np.concatenate([np.squeeze(t_for_ch) for t_for_ch in np.split(_temps, _temps.shape[2], axis=2)])
        else:
            return np.concatenate([np.squeeze(t) for t in np.split(templates_per_ch, templates_per_ch.shape[2], axis=2)])

    def _calculate_normalization_factor_per_component(self, templates_per_ch: np.ndarray) -> np.ndarray:
        _norm: np.ndarray = np.sum(templates_per_ch, axis=(0, 2))
        assert len(_norm.shape) == 1, (_norm.shape, self.number_of_components)
        assert len(_norm) == self.number_of_components, (len(_norm), _norm.shape, self.number_of_components)
        return _norm

    def generate_pure_templates(self) -> None:
        logger.info("Generating Pure Templates")
        if self._pure_mc_templates is not None:
            raise RuntimeError("Pure template shapes are already defined...")

        _templates = self._get_reco_ch_concatenated_base_templates(
            templates_per_ch=self.base_mc_bin_counts,
            normalize=False,
        )

        self._check_shape_dim_zero(a=_templates, after_zero_bin_mask=False, before_ch_concat=False)

        if self.fit_setup_info.remove_empty_bins:
            _templates = np.delete(_templates, self.empty_bin_mask, axis=0)

        self._check_shape_dim_one(a=_templates, is_single_dim=False)
        self._check_shape_dim_zero(a=_templates, after_zero_bin_mask=True, before_ch_concat=False)

        self._pure_mc_templates = _templates

    def check_base_mc_bin_counts(self, a: np.ndarray) -> None:
        assert not np.any(np.isnan(a)), np.sum(np.isnan(a))
        self._check_shape_dim_zero(a=a, after_zero_bin_mask=False, before_ch_concat=True)
        self._check_shape_dim_one(a=a, is_single_dim=False)
        self._check_shape_dim_two(a=a, has_less_dim=False)

    def generate_empty_bin_mask(self) -> None:
        if self.fit_setup_info.remove_empty_bins:
            _concatenated_bin_counts = self._get_reco_ch_concatenated_base_templates(
                templates_per_ch=self.base_mc_bin_counts,
                normalize=False,
            )
            # TODO: Should also be possible to set a lower bound of events, e.g. 20
            # TODO: Should also be possible to be based on data bin count, if data is used!!!
            _empty_bin_mask: np.ndarray = np.equal(np.sum(_concatenated_bin_counts, axis=1), 0.0)
        else:
            _empty_bin_mask = np.full(self.base_mc_bin_counts.shape[0], fill_value=False)

        assert not np.any(np.isnan(_empty_bin_mask)), np.sum(np.isnan(_empty_bin_mask))

        self._check_shape_dim_zero(a=_empty_bin_mask, after_zero_bin_mask=False, before_ch_concat=False)
        self._check_shape_dim_one(a=_empty_bin_mask, is_single_dim=True)
        self._empty_bin_mask = _empty_bin_mask

    @property
    def base_mc_bin_counts(self) -> np.ndarray:
        if self._base_mc_bin_counts is None:
            self._base_mc_bin_counts = self._calculate_base_mc_bin_counts(for_uncertainties=False)
            self.check_base_mc_bin_counts(a=self._base_mc_bin_counts)
        return self._base_mc_bin_counts

    @property
    def base_mc_bin_counts_for_uncertainties(self) -> np.ndarray:
        if self._base_mc_bin_counts_for_uncerts is None:
            self._base_mc_bin_counts_for_uncerts = self._calculate_base_mc_bin_counts(for_uncertainties=True)
            self.check_base_mc_bin_counts(a=self._base_mc_bin_counts)
        return self._base_mc_bin_counts_for_uncerts

    @property
    def squared_template_stat_error(self) -> np.ndarray:
        if self._squared_template_stat_error is None:
            self._squared_template_stat_error = self.calculate_squared_mc_stat_errors(for_templates=True)
        return self._squared_template_stat_error

    @property
    def template_shapes(self) -> np.ndarray:
        """
        Provides normalized base templates without any systematic effect on shapes!
        Systematic effects must be considered separately and based on the information this property provides.
        """
        if self._template_shapes is None:
            self.calculate_base_template_shapes()
        return self._template_shapes

    @property
    def pure_mc_templates(self) -> np.ndarray:
        if self._pure_mc_templates is None:
            self.generate_pure_templates()
        return self._pure_mc_templates

    @property
    def empty_bin_mask(self) -> np.ndarray:
        if self._empty_bin_mask is None:
            self.generate_empty_bin_mask()
        return self._empty_bin_mask

    # endregion

    # region Data bin count
    @property
    def data_df_weights(self) -> np.ndarray:
        if self._data_df_weights is None:
            assert self._data_df is not None

            if self._data_weight_col_name is None:
                self._data_df_weights = np.ones(len(self._data_df))
            else:
                self._data_df_weights = self._data_df[self._data_weight_col_name].values

        return self._data_df_weights

    def data_bin_counts(self) -> np.ndarray:
        if self._data_bin_counts is None:
            assert self._data_df is not None

            _bin_counts_per_ch: List[np.ndarray] = []

            for reco_ch in self.fit_setup_info.reco_channels:
                _channel_padding: Tuple[int, int] = self._get_padding_for_channel(reco_ch_info=reco_ch)

                reco_ch_mask: np.ndarray = reco_ch.get_mask(
                    df=self._data_df,
                    reco_mode_column_str=self.fit_setup_info.reco_mode_col,
                )

                _bin_count = np.histogramdd(
                    sample=self._data_df[reco_ch_mask][list(self.obs_column_names)].values,
                    weights=self.data_df_weights[reco_ch_mask],
                    bins=reco_ch.n_bins_per_obs,
                    range=reco_ch.scope_per_obs,
                )[0]

                _bin_counts_per_ch.append(
                    np.pad(_bin_count.flatten(), pad_width=_channel_padding, mode="constant", constant_values=0.0)
                    if _channel_padding[1] > 0
                    else _bin_count.flatten()
                )

            _data_bin_counts: np.ndarray = np.concatenate(_bin_counts_per_ch, axis=0)

            self._check_shape_dim_one(a=_data_bin_counts, is_single_dim=True)
            self._check_shape_dim_zero(a=_data_bin_counts, after_zero_bin_mask=False, before_ch_concat=False)

            if sum(self.empty_bin_mask):
                assert len(self.empty_bin_mask) == _data_bin_counts.shape[0], (
                    len(self.empty_bin_mask),
                    _data_bin_counts.shape,
                )
                _data_bin_counts = np.delete(_data_bin_counts, self.empty_bin_mask, axis=0)

            self._check_shape_dim_zero(a=_data_bin_counts, after_zero_bin_mask=True, before_ch_concat=False)

            self._data_bin_counts = _data_bin_counts

        return self._data_bin_counts

    def calculate_data_bin_stat_errors_squared(self) -> np.ndarray:
        assert self._data_df is not None

        _bin_errors_squared_per_ch: List[np.ndarray] = []

        for reco_ch in self.fit_setup_info.reco_channels:
            _channel_padding: Tuple[int, int] = self._get_padding_for_channel(reco_ch_info=reco_ch)
            reco_ch_mask: np.ndarray = reco_ch.get_mask(
                df=self._data_df,
                reco_mode_column_str=self.fit_setup_info.reco_mode_col,
            )

            _bin_errors_squared = np.histogramdd(
                sample=self._data_df[reco_ch_mask][list(self.obs_column_names)].values,
                weights=np.square(self.data_df_weights[reco_ch_mask]),
                bins=reco_ch.n_bins_per_obs,
                range=reco_ch.scope_per_obs,
            )[0]

            _bin_errors_squared_per_ch.append(
                np.pad(_bin_errors_squared.flatten(), pad_width=_channel_padding, mode="constant", constant_values=0.0)
                if _channel_padding[1] > 0
                else _bin_errors_squared.flatten()
            )

        _data_bin_errors_squared: np.ndarray = np.concatenate(_bin_errors_squared_per_ch, axis=0)

        self._check_shape_dim_zero(a=_data_bin_errors_squared, after_zero_bin_mask=False, before_ch_concat=False)
        self._check_shape_dim_one(a=_data_bin_errors_squared, is_single_dim=True)

        if sum(self.empty_bin_mask):
            _data_bin_errors_squared = np.delete(_data_bin_errors_squared, self.empty_bin_mask, axis=0)

        self._check_shape_dim_zero(a=_data_bin_errors_squared, after_zero_bin_mask=True, before_ch_concat=False)

        return _data_bin_errors_squared

    @property
    def data_bin_stat_errors_squared(self) -> np.ndarray:
        if self._data_bin_stat_errors_squared is None:
            self._data_bin_stat_errors_squared = self.calculate_data_bin_stat_errors_squared()
        return self._data_bin_stat_errors_squared

    @property
    def data_bin_stat_errors(self) -> np.ndarray:
        return np.sqrt(self.data_bin_stat_errors_squared)

    # endregion

    # region Local efficiencies
    def calculate_local_efficiency_infos(
        self,
        df: pd.DataFrame,
        reco_ch_infos: List[Tuple[RecoChannelInfo, np.ndarray]],
        combined_mask: np.ndarray,
    ) -> None:
        logger.info("In initialization of mc_df: calculating efficiency information")

        for comp in self.fit_setup_info.components:
            comp_mask: np.ndarray = comp.get_component_mask(
                df=df,
                component_id_column_name=self.fit_setup_info.component_id_column_name,
            )

            _wo_sel_for_this_comp_count: float = float(np.sum(df[comp_mask][self.mc_weight_col_name]))
            _wo_sel_for_this_comp_count_uncertainty: float = np.sqrt(
                np.sum(np.square(df[comp_mask][self.mc_weight_col_name]))
            )

            all_for_this_comp_count: float = float(np.sum(df[comp_mask & combined_mask][self.mc_weight_col_name]))
            all_for_this_comp_count_uncertainty: float = np.sqrt(
                np.sum(np.square(df[comp_mask & combined_mask][self.mc_weight_col_name]))
            )

            for reco_ch_info, reco_ch_mask in reco_ch_infos:

                comp_count: float = float(np.sum(df[comp_mask & reco_ch_mask][self.mc_weight_col_name]))
                comp_count_uncertainty: float = np.sqrt(
                    np.sum(np.square(df[comp_mask & reco_ch_mask][self.mc_weight_col_name]))
                )

                self._local_efficiency_infos.append(
                    LocalEfficiencyInfo(
                        component_id=comp.comp_index,
                        reco_channel_id=reco_ch_info.ch_index,
                        n_before=ufloat(all_for_this_comp_count, all_for_this_comp_count_uncertainty),
                        n_after=ufloat(comp_count, comp_count_uncertainty),
                        n_before_selection=ufloat(_wo_sel_for_this_comp_count, _wo_sel_for_this_comp_count_uncertainty),
                    )
                )

        self._local_efficiency_infos.check_consistency()

    def calculate_local_efficiencies(self) -> np.ndarray:  # TODO: Make local efficiencies floating?
        local_effs_per_comp: List[np.ndarray] = []

        for comp in self.fit_setup_info.components:
            _local_effs_per_reco_ch: List[np.ndarray] = [
                np.full(
                    self.fit_setup_info.max_n_flattened_bins_per_ch,
                    fill_value=self._local_efficiency_infos[(comp.comp_index, ch.ch_index)].efficiency,
                )
                for ch in self.fit_setup_info.reco_channels
            ]

            assert len(self.empty_bin_mask) == sum(len(a) for a in _local_effs_per_reco_ch), (
                len(self.empty_bin_mask),
                sum(len(a) for a in _local_effs_per_reco_ch),
            )
            local_effs_per_comp.append(np.concatenate(_local_effs_per_reco_ch)[~self.empty_bin_mask])

            assert len(local_effs_per_comp[-1]) == sum(~self.empty_bin_mask), (
                len(local_effs_per_comp[-1]),
                sum(~self.empty_bin_mask),
            )

        _local_effs: np.ndarray = np.stack(local_effs_per_comp, axis=1)

        assert self.number_of_components == _local_effs.shape[1], (self.number_of_components, _local_effs.shape)
        assert self.final_n_bins_total == _local_effs.shape[0], (self.final_n_bins_total, _local_effs.shape)

        return _local_effs

    @property
    def local_efficiencies(self) -> np.ndarray:
        if self._local_efficiencies is None:
            self._local_efficiencies = self.calculate_local_efficiencies()
        return self._local_efficiencies

    # endregion

    # region Global efficiency input

    @property
    def global_eff_info(self) -> SimpleEfficiencyCalculator:
        if self._global_eff_info is None:
            self._global_eff_info = self.generate_global_efficiency_infos()
        return self._global_eff_info

    def generate_global_efficiency_infos(self) -> SimpleEfficiencyCalculator:
        if self.file_dict_dir_path is None:
            raise RuntimeError(f"The fitter {self.__class__.__name__} requires the argument 'file_dict_dir_path'!")

        return SimpleEfficiencyCalculator(
            df=self._mc_df,
            sig_id_column_name=self.fit_setup_info.component_id_column_name,
            sig_id_data_type_mapping=self.fit_setup_info.sig_id_to_data_type_mapping,
            file_dict_base_path=self.file_dict_dir_path,
            use_new_fei_calibration_weights=False,
            additional_scaling_weight_columns=None,
        )

    # endregion

    # region Parameters

    @property
    def initial_parameters(self) -> np.ndarray:
        return self._initial_params_vector

    def set_initial_parameters(self, values: np.ndarray) -> None:
        assert self._initial_params_vector == np.array([-999]), (self._initial_params_vector, values)
        self._initial_params_vector = values

    @property
    def parameters(self) -> np.ndarray:
        return self._params_vector

    def set_parameters(self, values: np.ndarray) -> None:
        assert self._params_vector == np.array([-999]), (self._params_vector, values)
        self._params_vector = values

    # endregion

    # region Abstract Cost function methods
    @abstractmethod
    def cost_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'cost_fcn' must be implemented in the derived class.")

    @abstractmethod
    def chi2_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'chi2_fcn' must be implemented in the derived class.")

    @abstractmethod
    def gaussian_constrains(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'gaussian_constrains_fcn' must be implemented in the derived class.")

    @abstractmethod
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        raise NotImplementedError("The 'get_yields' must be implemented in the derived class.")

    # endregion

    # region Methods for post-processing and evaluation of results
    @classmethod
    def parameter_name_mapping(cls) -> Dict[int, str]:
        return {i: n for i, n in enumerate(cls.ordered_parameter_names)}

    # endregion

    # region Utility functions
    @staticmethod
    def pop_kwarg_with_default(
        kw_args: Dict[str, Any],
        key: str,
        default_value: Optional[Any] = None,
    ) -> Any:
        if key in kw_args:
            return kw_args.pop(key)
        else:
            return default_value

    def set_iminuit_properties(
        self,
        minuit_instance: Minuit,
    ) -> None:
        minuit_instance.errors = self.minuit_settings.error_factor * self.initial_parameters
        minuit_instance.errordef = self.minuit_settings.error_def
        minuit_instance.print_level = self.minuit_settings.print_level

    # endregion

    # region Parameter limits
    @property
    def parameter_limits(self) -> Tuple[Tuple[int, Tuple[Optional[float], Optional[float]]], ...]:
        parameter_limits = getattr(self, "_parameter_limits", None)
        if parameter_limits is None:
            return tuple()

        assert isinstance(parameter_limits, (tuple, list, set))
        assert all(isinstance(e, tuple) for e in parameter_limits)
        assert all(len(e) == 2 for e in parameter_limits)
        assert all(isinstance(i, int) for i, _ in parameter_limits)
        assert all(isinstance(lims, tuple) for _, lims in parameter_limits)
        assert all(len(lims) == 2 for _, lims in parameter_limits)
        assert all(all((isinstance(e, float) or e is None) for e in lims) for _, lims in parameter_limits)
        return tuple(parameter_limits)

    # endregion

    # region Fit result evaluation
    def get_rescaled_mc_sample_with(
        self,
        sample: pd.DataFrame,
        weight_col_name: str,
        parameters: np.ndarray,
        add_required_sig_id_col: bool,
    ) -> pd.DataFrame:
        assert weight_col_name in sample.columns, weight_col_name
        _sample: pd.DataFrame = copy.deepcopy(sample)
        if add_required_sig_id_col:
            self.fit_setup_info.prepare_mc_df(df=_sample)

        _ones: np.ndarray = np.ones(shape=len(_sample))
        _test_zero: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        initial_yields: np.ndarray = self.get_yields(params=self.initial_parameters)
        fitted_yields: np.ndarray = self.get_yields(params=parameters)
        ratio: np.ndarray = fitted_yields / initial_yields

        assert len(self.fit_setup_info.components) == len(ratio), (len(self.fit_setup_info.components), len(ratio))

        for ratio_index, component in enumerate(self.fit_setup_info.components):
            comp_mask: np.ndarray = component.get_component_mask(
                df=_sample,
                component_id_column_name=self.fit_setup_info.component_id_column_name,
            )

            assert not any(_test_zero & comp_mask), (ratio_index, component.name)
            _test_zero |= comp_mask

            _sample.loc[comp_mask, weight_col_name] *= _ones[comp_mask] * ratio[ratio_index]

        return _sample

    # endregion
