"""
Tools for rdstar fit binning plots results.

Felix Metzner 2023
"""


import os
import numpy as np
import pandas as pd
import seaborn as sns

from matplotlib import pyplot as plt
from typing import Tuple, Dict, Union, ClassVar

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_utilities import export, AxesType, FigureType

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeIDInfo, RecoModeGroupInfo

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.rdstar_fit_obs_setups import FitObsSetupMain
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import base_rdstar_weight_info


__all__ = [
    "RDStarFitBinningPlots",
]


logger = log_to_default(name=__name__)


BinningInfoElementType = Union[np.ndarray, Tuple[float, ...]]
BinningInfoType = Tuple[BinningInfoElementType, BinningInfoElementType]


# region plot infos
fit_obs_mms_hist_var: HistVariable = FitObsSetupMain.obs2.hist_var
fit_obs_neextra_hist_var: HistVariable = FitObsSetupMain.obs1.hist_var

# endregion


# region RDStar fit binning plotter class
class RDStarFitBinningPlots:

    binning_plots_dir_name: ClassVar[str] = "RDStarBinningPlots"
    binning_plots_name: ClassVar[str] = "binning_heatmap_plot"

    def __init__(
        self,
        evaluation_name: str,
        base_output_path: PathType,
    ) -> None:

        self.eval_name: str = evaluation_name
        self.base_output_path: PathType = base_output_path

        self.output_path: PathType = os.path.join(self.base_output_path, self.binning_plots_dir_name, self.eval_name)

        self.binning_info_dict: Dict[str, BinningInfoType] = FitObsSetupMain.binning_for_reco_modes

    def produce_fit_binning_plots(
        self,
        df: pd.DataFrame,
        is_mc: bool,
    ) -> None:
        binning_plots_path: PathType = os.path.join(self.output_path, "MC" if is_mc else "Data")
        os.makedirs(binning_plots_path, exist_ok=True)

        for reco_mode_info_col in BSigRecoModes.main_modes.reco_mode_infos:
            _ = self.plot_bin_counts_for_binning(
                df=df,
                is_mc=is_mc,
                reco_modes_info=reco_mode_info_col,
                binning_plots_path=binning_plots_path,
            )

    @staticmethod
    def check_binning_info(binning_info: BinningInfoType) -> None:
        obs_1_binning: BinningInfoElementType = binning_info[0]
        obs_2_binning: BinningInfoElementType = binning_info[1]
        assert obs_1_binning[0] == fit_obs_mms_hist_var.scope[0], (obs_1_binning, fit_obs_mms_hist_var.scope)
        assert obs_1_binning[-1] == fit_obs_mms_hist_var.scope[1], (obs_1_binning, fit_obs_mms_hist_var.scope)

        assert obs_2_binning[0] == fit_obs_neextra_hist_var.scope[0], (obs_2_binning, fit_obs_neextra_hist_var.scope)
        assert obs_2_binning[-1] == fit_obs_neextra_hist_var.scope[1], (obs_2_binning, fit_obs_neextra_hist_var.scope)

    @staticmethod
    def get_evt_count_for(
        df: pd.DataFrame,
        base_mask: np.ndarray,
        binning_info: BinningInfoType,
        is_mc: bool,
    ) -> np.ndarray:
        var1_mask: np.ndarray = (df[fit_obs_mms_hist_var.df_label].values >= fit_obs_mms_hist_var.scope[0]) & (
            df[fit_obs_mms_hist_var.df_label].values <= fit_obs_mms_hist_var.scope[1]
        )
        var2_mask: np.ndarray = (df[fit_obs_neextra_hist_var.df_label].values >= fit_obs_neextra_hist_var.scope[0]) & (
            df[fit_obs_neextra_hist_var.df_label].values <= fit_obs_mms_hist_var.scope[1]
        )

        mask: np.ndarray = base_mask & var1_mask & var2_mask

        if is_mc:
            weights_array = df[mask][base_rdstar_weight_info.col_name].values
        else:
            weights_array = np.ones(df[mask][fit_obs_mms_hist_var.df_label].shape)

        evt_count, _ = np.histogramdd(
            sample=df[mask][[fit_obs_mms_hist_var.df_label, fit_obs_neextra_hist_var.df_label]].values,
            weights=weights_array,
            bins=binning_info,
            range=None,
            density=False,
        )
        return evt_count

    def plot_bin_counts_for_binning(
        self,
        df: pd.DataFrame,
        is_mc: bool,
        reco_modes_info: Union[RecoModeIDInfo, RecoModeGroupInfo],
        binning_plots_path: PathType,
    ) -> Tuple[FigureType, AxesType]:
        binning_info: BinningInfoType = FitObsSetupMain.binning_for_reco_modes[reco_modes_info.name]
        self.check_binning_info(binning_info=binning_info)

        reco_ids: Tuple[int, ...] = reco_modes_info.ids
        base_mask: np.ndarray = df[Cols.decay_mode_column_name].isin(reco_ids).to_numpy()
        reco_mode_latex_label: str = reco_modes_info.latex_label

        bin_counts: np.ndarray = self.get_evt_count_for(
            df=df,
            base_mask=base_mask,
            binning_info=binning_info,
            is_mc=is_mc,
        )

        fig: FigureType = plt.figure(num=None, figsize=(6, 4.5), dpi=200, facecolor="w", edgecolor="k")
        ax: AxesType = plt.gca()

        sns.heatmap(
            data=bin_counts,
            annot=True,
            cmap="crest",
            ax=ax,
            annot_kws={"fontsize": 6},
        )

        title_start: str = "MC" if is_mc else "Data"
        ax.set_title(title_start + " in " + reco_mode_latex_label, loc="left")
        ax.set_xlabel(fit_obs_neextra_hist_var.variable_name)
        ax.set_ylabel(fit_obs_mms_hist_var.variable_name)

        nb_d1: int = (binning_info[0].shape[0] if isinstance(binning_info[0], np.ndarray) else len(binning_info[0])) - 1
        nb_d2: int = (binning_info[1].shape[0] if isinstance(binning_info[1], np.ndarray) else len(binning_info[1])) - 1

        binning_text: str = rf"{fit_obs_mms_hist_var.variable_name} ({nb_d1} bins): {list(binning_info[0])}"
        binning_text += "\n" + rf"{fit_obs_neextra_hist_var.variable_name} ({nb_d2} bins): {list(binning_info[1])}"
        ax.text(
            x=0.04,
            y=-0.15,
            s=binning_text,
            transform=ax.transAxes,
            va="top",
            fontsize="small",
        )

        export(
            fig=fig,
            filename=f"{self.binning_plots_name}_in_{reco_modes_info.name}",
            target_dir=binning_plots_path,
            close_figure=False,
        )

        return fig, ax


# endregion
