"""
Collection of B tagging calibration fitters.

Felix Metzner 2023
"""

from typing import Tuple, Dict, Type

from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.reco_modes.reco_mode_infos import U4SRecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters.tag_cal_fitter_base import TagCorrFitterBase

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_main
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_main_two_bkg

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_one_bkg_floating_df
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_one_bkg_coupled_df
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_one_bkg_coupled_d_ds
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_one_bkg_extra_df_corr

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import (
    fitter_tag_cal_combined_bkg_floating_cf,
    fitter_tag_cal_combined_bkg_floating_df,
    fitter_tag_cal_combined_bkg_coupled_df,
    fitter_tag_cal_combined_bkg_coupled_d_ds,
    fitter_tag_cal_combined_bkg_extra_df_corr,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_split_bkg_floating_cf
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_split_bkg_floating_df
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_split_bkg_coupled_df
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_split_bkg_coupled_d_ds

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import (
    fitter_tag_cal_per_tag_mode_combined_bkg_coupled_df,
    fitter_tag_cal_per_tag_mode_combined_bkg_floating_ds,
    fitter_tag_cal_per_tag_mode_charge_split_coupled_d,
    fitter_tag_cal_per_tag_mode_charge_split_floating_ds,
    fitter_tag_cal_per_tag_mode_b_charge_coupled_d,
    fitter_tag_cal_per_tag_mode_b_charge_floating_ds,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_sig_tag_closure_a
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import fitter_tag_cal_sig_tag_closure_b


__all__ = [
    "tc_fitter_dict",
    "keys_to_evaluate",
    "right_sig_tag_combination_selection",
    "wrong_sig_tag_combination_selection",
]


right_sig_tag_combination_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{U4SRecoModeInfos.u4s_from_neutral_reco_mode.id} <= {U4SRecoModeInfos.u4s_reco_mode_col_name} <= {U4SRecoModeInfos.u4s_from_mixed_neutral_reco_mode.id}",
    cut_latex_label=rf"Matching {LatexObsNames.reco_y4s_reco_mode_label}s",
    cut_var_col_name=U4SRecoModeInfos.u4s_reco_mode_col_name,
)

wrong_sig_tag_combination_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{U4SRecoModeInfos.u4s_from_neutral_wt_reco_1_mode.id} <= {U4SRecoModeInfos.u4s_reco_mode_col_name} <= {U4SRecoModeInfos.u4s_from_charged_wt_reco_2_mode.id}",
    cut_latex_label=rf"Wrong-Tag {LatexObsNames.reco_y4s_reco_mode_label}s",
    cut_var_col_name=U4SRecoModeInfos.u4s_reco_mode_col_name,
)


tc_fitter_dict: Dict[str, Type[TagCorrFitterBase]] = {
    # Used Tag Cal Setups:
    "tc_main": fitter_tag_cal_main.TagCorrectionFitter,
    "tc_main_2b": fitter_tag_cal_main_two_bkg.TagCorrectionFitter,
    # "tc_dss": fitter_tag_cal_dss.TagCorrectionFitter,
    # One BKG Component including B -> D l nu UP FEED into BKG for B -> D* l reco channels
    "tc_one_bkg_floating_df": fitter_tag_cal_one_bkg_floating_df.TagCorrectionFitter,
    "tc_one_bkg_coupled_df": fitter_tag_cal_one_bkg_coupled_df.TagCorrectionFitter,
    "tc_one_bkg_coupled_d_ds": fitter_tag_cal_one_bkg_coupled_d_ds.TagCorrectionFitter,
    "tc_one_bkg_extra_df_corr": fitter_tag_cal_one_bkg_extra_df_corr.TagCorrectionFitter,
    # Combined BKG Component WITHOUT B -> D l nu UP FEED
    "tc_combined_bkg_floating_cf": fitter_tag_cal_combined_bkg_floating_cf.TagCorrectionFitter,
    "tc_combined_bkg_floating_df": fitter_tag_cal_combined_bkg_floating_df.TagCorrectionFitter,
    "tc_combined_bkg_coupled_df": fitter_tag_cal_combined_bkg_coupled_df.TagCorrectionFitter,
    "tc_combined_bkg_coupled_d_ds": fitter_tag_cal_combined_bkg_coupled_d_ds.TagCorrectionFitter,
    "tc_combined_bkg_extra_df_corr": fitter_tag_cal_combined_bkg_extra_df_corr.TagCorrectionFitter,
    # Split BKG Components
    "tc_split_bkg_floating_cf": fitter_tag_cal_split_bkg_floating_cf.TagCorrectionFitter,
    "tc_split_bkg_floating_df": fitter_tag_cal_split_bkg_floating_df.TagCorrectionFitter,
    "tc_split_bkg_coupled_df": fitter_tag_cal_split_bkg_coupled_df.TagCorrectionFitter,
    "tc_split_bkg_coupled_d_ds": fitter_tag_cal_split_bkg_coupled_d_ds.TagCorrectionFitter,
    #
    # FEI Tag Calibration per B-Tag Reco-Modes
    "tc_ptm_combined_bkg_coupled_df": fitter_tag_cal_per_tag_mode_combined_bkg_coupled_df.TagCorrectionFitter,
    "tc_ptm_combined_bkg_floating_ds": fitter_tag_cal_per_tag_mode_combined_bkg_floating_ds.TagCorrectionFitter,
    "tc_ptm_charge_split_coupled_d": fitter_tag_cal_per_tag_mode_charge_split_coupled_d.TagCorrectionFitter,
    "tc_ptm_charge_split_floating_ds": fitter_tag_cal_per_tag_mode_charge_split_floating_ds.TagCorrectionFitter,
    "tc_ptm_b_charge_coupled_d": fitter_tag_cal_per_tag_mode_b_charge_coupled_d.TagCorrectionFitter,
    "tc_ptm_b_charge_floating_ds": fitter_tag_cal_per_tag_mode_b_charge_floating_ds.TagCorrectionFitter,
    #
    # FEI Tag Calibration Closure Tests
    "tc_closure_test_a": fitter_tag_cal_sig_tag_closure_a.TagCorrectionFitter,
    "tc_closure_test_b": fitter_tag_cal_sig_tag_closure_b.TagCorrectionFitter,
}


keys_to_evaluate: Tuple[str, ...] = (
    "tc_main",
    # "tc_main_2b",
    # "tc_one_bkg_floating_df",
    # "tc_one_bkg_coupled_df",
    # "tc_one_bkg_coupled_d_ds",
    # "tc_combined_bkg_floating_df",
    # "tc_combined_bkg_coupled_df",
    # "tc_combined_bkg_coupled_d_ds",
    # "tc_ptm_b_charge_coupled_d",
    # "tc_ptm_b_charge_floating_ds",
    # "tc_ptm_combined_bkg_coupled_df",
    # "tc_ptm_combined_bkg_floating_ds",
    # "tc_ptm_charge_split_coupled_d",
    # "tc_ptm_charge_split_floating_ds",
    # "tc_closure_test_a",
    # "tc_closure_test_b",
)


assert all(k in tc_fitter_dict for k in keys_to_evaluate), [k for k in keys_to_evaluate if k not in tc_fitter_dict]
