"""
Generator level MC matching information.

Felix Metzner 2023
"""

from typing import Tuple, Dict, Type, Any

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo

from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_queries import GenMCMatchingQueries as GenMCQ
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_utilities import (
    GenQueryDictType,
    get_gen_b_id_col_name,
)


__all__ = [
    "GenMCIDs",
    "GenMCMatchingInfo",
]


class GenMCIDs:
    Dtau: int = 1
    Dsttau: int = 2
    Dl: int = 3
    Dstl: int = 4
    Donel: int = 51
    Dtwostarl: int = 52
    Dprimeonel: int = 53
    Dzerostarl: int = 54
    Donetau: int = 55
    Dtwostartau: int = 56
    Dprimeonetau: int = 57
    Dzerostartau: int = 58
    hadronic_tau: int = 6
    other: int = 7
    continuum: int = 8
    PiLNu: int = 9

    @staticmethod
    def get_gen_ids_dict() -> Dict[str, int]:
        _method_names: Tuple[str, ...] = (
            "get_gen_ids_dict",
            "get_all_gen_ids",
            "run_checks",
        )

        return {
            attr_name: val
            for attr_name, val in GenMCIDs.__dict__.items()
            if not ((attr_name in _method_names) or (attr_name.endswith("__") or attr_name.startswith("_")))
        }

    @staticmethod
    def get_all_gen_ids() -> Tuple[int, ...]:
        return tuple(gen_id for gen_id in GenMCIDs.get_gen_ids_dict().values())

    @staticmethod
    def run_checks() -> None:
        __gen_id_dict: Dict[str, int] = GenMCIDs.get_gen_ids_dict()

        assert all(isinstance(k, str) for k in __gen_id_dict.keys()), [
            (k, type(k)) for k in __gen_id_dict.keys() if not isinstance(k, str)
        ]
        assert all(isinstance(v, int) for v in __gen_id_dict.values()), [
            (k, v, type(v)) for k, v in __gen_id_dict.items() if not isinstance(v, int)
        ]

        __all_gen_ids: Tuple[int, ...] = GenMCIDs.get_all_gen_ids()

        assert len(set(__all_gen_ids)) == len(__all_gen_ids), (
            len(__all_gen_ids),
            len(set(__all_gen_ids)),
            set([i for i in set(__all_gen_ids) if __all_gen_ids.count(i) > 1]),
        )


class GenMCMatchingInfo:
    gen_b1_id_col_name: str = get_gen_b_id_col_name(b_index=1)
    gen_b2_id_col_name: str = get_gen_b_id_col_name(b_index=2)

    to_inconclusive_gen_mc_id_flag: int = -999
    to_inconclusive_gen_mc_id_name: str = "NO_GEN_MC_MATCH"

    required_columns: Tuple[str, ...] = GenMCQ.required_columns
    gen_ids_definition_class: Type[GenMCIDs] = GenMCIDs

    # fmt: off
    Dtau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dtau, query=GenMCQ.b_to_d_tau, latex_label=LatexDecayProcesses.to_d_tau_nu)
    Dsttau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dsttau, query=GenMCQ.b_to_dst_tau, latex_label=LatexDecayProcesses.to_d_st_tau_nu)
    Dl: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dl, query=GenMCQ.b_to_d_l, latex_label=LatexDecayProcesses.to_d_l_nu)
    Dstl: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dstl, query=GenMCQ.b_to_d_st_l, latex_label=LatexDecayProcesses.to_d_st_l_nu)
    Donel: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Donel, query=GenMCQ.b_to_d_one_l, latex_label=LatexDecayProcesses.to_d_1_l_nu)
    Dtwostarl: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dtwostarl, query=GenMCQ.b_to_d_twostar_l, latex_label=LatexDecayProcesses.to_dst_2_l_nu)
    Dprimeonel: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dprimeonel, query=GenMCQ.b_to_d_primeone_l, latex_label=LatexDecayProcesses.to_dprime1_l_nu)
    Dzerostarl: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dzerostarl, query=GenMCQ.b_to_d_zerostar_l, latex_label=LatexDecayProcesses.to_d0st_l_nu)
    Donetau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Donetau, query=GenMCQ.b_to_d_one_tau, latex_label=LatexDecayProcesses.to_d_1_tau_nu)
    Dtwostartau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dtwostartau, query=GenMCQ.b_to_d_twostar_tau, latex_label=LatexDecayProcesses.to_dst_2_tau_nu)
    Dprimeonetau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dprimeonetau, query=GenMCQ.b_to_d_primeone_tau, latex_label=LatexDecayProcesses.to_dprime1_tau_nu)
    Dzerostartau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.Dzerostartau, query=GenMCQ.b_to_d_zerostar_tau, latex_label=LatexDecayProcesses.to_d0st_tau_nu)
    hadronic_tau: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.hadronic_tau, query=GenMCQ.hadronic_tau, latex_label=LatexDecayProcesses.to_hadronic_tau_bkg_txt)
    other: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.other, query=GenMCQ.other, latex_label=LatexDecayProcesses.to_other_bkg_txt)
    continuum: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.continuum, query=GenMCQ.continuum, latex_label=LatexDecayProcesses.to_continuum_txt)
    PiLNu: MatchQueryInfo = MatchQueryInfo(id_info=GenMCIDs.PiLNu, query=GenMCQ.b_to_pi_l, latex_label=LatexDecayProcesses.to_pi_l_nu)
    # fmt: on

    @staticmethod
    def get_gen_mc_id_query_dict() -> GenQueryDictType:
        _member_names: Tuple[str, ...] = (
            "gen_b1_id_col_name",
            "gen_b2_id_col_name",
            "to_inconclusive_gen_mc_id_flag",
            "to_inconclusive_gen_mc_id_name",
            "required_columns",
            "gen_ids_definition_class",
        )

        _method_names: Tuple[str, ...] = (
            "get_gen_mc_id_query_dict",
            "get_all_gen_ids",
            "get_gen_id_queries",
            "get_gen_id_mapping",
            "gen_id_to_latex_label",
            "gen_id_name_to_latex_label",
            "run_checks",
        )

        _query_dict: Dict[str, Any] = {
            attr_name: val
            for attr_name, val in GenMCMatchingInfo.__dict__.items()
            if not ((attr_name in _member_names + _method_names) or attr_name.endswith("__") or attr_name.startswith("_"))
        }
        assert all(isinstance(val, MatchQueryInfo) for val in _query_dict.values()), [
            (k, type(v)) for k, v in _query_dict.items() if not isinstance(v, MatchQueryInfo)
        ]
        query_dict: Dict[str, MatchQueryInfo] = {k: v for k, v in _query_dict.items() if isinstance(v, MatchQueryInfo)}
        return query_dict

    @staticmethod
    def get_all_gen_ids() -> Tuple[int, ...]:
        return tuple(v.id for v in GenMCMatchingInfo.get_gen_mc_id_query_dict().values())

    @staticmethod
    def get_gen_id_queries() -> Tuple[MatchQueryInfo, ...]:
        return tuple(gen_id_info for gen_id_info in GenMCMatchingInfo.get_gen_mc_id_query_dict().values())

    @staticmethod
    def get_gen_id_mapping() -> GenQueryDictType:
        return GenMCMatchingInfo.get_gen_mc_id_query_dict()

    @staticmethod
    def gen_id_to_latex_label() -> Dict[int, str]:
        _map: Dict[int, str] = {v.id: v.latex_label for v in GenMCMatchingInfo.get_gen_mc_id_query_dict().values()}
        _map.update({GenMCMatchingInfo.to_inconclusive_gen_mc_id_flag: LatexDecayProcesses.to_inconclusive_txt})
        return _map

    @staticmethod
    def gen_id_name_to_latex_label() -> Dict[str, str]:
        _map: Dict[str, str] = {k: v.latex_label for k, v in GenMCMatchingInfo.get_gen_mc_id_query_dict().items()}
        _map.update({GenMCMatchingInfo.to_inconclusive_gen_mc_id_name: LatexDecayProcesses.to_inconclusive_txt})
        return _map

    @staticmethod
    def run_checks() -> None:
        __all_gen_ids_in_query_info: Tuple[int, ...] = GenMCMatchingInfo.get_all_gen_ids()

        assert len(__all_gen_ids_in_query_info) == len(set(__all_gen_ids_in_query_info)), (
            len(__all_gen_ids_in_query_info),
            len(set(__all_gen_ids_in_query_info)),
            set([i for i in set(__all_gen_ids_in_query_info) if __all_gen_ids_in_query_info.count(i) > 1]),
        )

        __all_defined_gen_ids: Tuple[int, ...] = GenMCMatchingInfo.gen_ids_definition_class.get_all_gen_ids()
        gen_id_def_class_name: str = GenMCMatchingInfo.gen_ids_definition_class.__name__

        assert all([i in __all_defined_gen_ids for i in __all_gen_ids_in_query_info]), (
            f"Gen IDs used in GenMCMatchingInfo that is NOT DEFINED in {gen_id_def_class_name} class:",
            [i for i in __all_gen_ids_in_query_info if i not in __all_defined_gen_ids],
        )

        assert all([i in __all_gen_ids_in_query_info for i in __all_defined_gen_ids]), (
            "Gen IDs defined in {gen_id_def_class_name} class that are NOT USED in GenMCMatchingInfo:",
            [i for i in __all_defined_gen_ids if i not in __all_gen_ids_in_query_info],
        )


original_gen_id_to_label_dict: Dict[int, str] = {
    1: LatexDecayProcesses.to_d_tau_nu,
    2: LatexDecayProcesses.to_d_st_tau_nu,
    3: LatexDecayProcesses.to_d_l_nu,
    4: LatexDecayProcesses.to_d_st_l_nu,
    51: LatexDecayProcesses.to_d_1_l_nu,
    52: LatexDecayProcesses.to_dst_2_l_nu,
    53: LatexDecayProcesses.to_dprime1_l_nu,
    54: LatexDecayProcesses.to_d0st_l_nu,
    55: LatexDecayProcesses.to_d_1_tau_nu,
    56: LatexDecayProcesses.to_dst_2_tau_nu,
    57: LatexDecayProcesses.to_dprime1_tau_nu,
    58: LatexDecayProcesses.to_d0st_tau_nu,
    6: LatexDecayProcesses.to_hadronic_tau_bkg_txt,
    7: LatexDecayProcesses.to_other_bkg_txt,
    8: LatexDecayProcesses.to_continuum_txt,
    9: LatexDecayProcesses.to_pi_l_nu,
    -999: LatexDecayProcesses.to_inconclusive_txt,
}

if __name__ == "__main__":
    # Checks for GenMCIDs
    GenMCIDs.run_checks()

    # Checks for GenMCMatchingInfo
    GenMCMatchingInfo.run_checks()

    for gen_id_key, gen_id_value in GenMCIDs.get_gen_ids_dict().items():
        print(f"{gen_id_key}: {gen_id_value}")

    new_gen_id_to_label_dict: Dict[int, str] = GenMCMatchingInfo.gen_id_to_latex_label()
    assert original_gen_id_to_label_dict == new_gen_id_to_label_dict, (
        new_gen_id_to_label_dict,
        original_gen_id_to_label_dict,
    )

    print(GenMCIDs.get_all_gen_ids())

    print(new_gen_id_to_label_dict)
