"""
This module provides the plotting methods for the rdstar analysis.

Felix Metzner 2019
"""

import os
import copy
import numpy as np
import pandas as pd
from scipy.stats import chi2
import matplotlib.pyplot as plt

from typing import Any, Union, Optional, Tuple, Dict, Generator

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import StackedHistogramPlot
from templatefitter.plotter.plot_utilities import FigureType, AxesType
from templatefitter.binned_distributions.binning import Binning, BinsInputType

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollections

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import (
    RecoModeCollection,
    RecoModeDictType,
    RecoModeGroupNameToTitleDictType,
)


__all__ = [
    "rdstar_color_map",
    #
    "default_sig_mc_matching_collection",
    #
    "dss_mc_reweighting_latex_titles",
    #
    "plot_chi_squared_toys",
    "plot_stacked_for_all_modes",
]


logger = log_to_default(name=__name__)


plot_style.set_matplotlibrc_params()


rdstar_color_map: str = "viridis"

default_sig_mc_matching_collection: MCMatchingCollection = MCMatchingCollections.original__base__collection


dss_mc_reweighting_latex_titles: RecoModeGroupNameToTitleDictType = {
    "Done": LatexDecayProcesses.b_to_d_1_l_nu,
    "Dtwostar": LatexDecayProcesses.b_to_dst_2_l_nu,
    "Dprimeone": LatexDecayProcesses.b_to_dprime1_l_nu,
    "Dzerostar": LatexDecayProcesses.b_to_d0st_l_nu,
}


def plot_stacked(
    df: pd.DataFrame,
    variable: HistVariable,
    weight_column: Optional[str] = None,
    sig_mc_matching_collection: Optional[MCMatchingCollection] = None,
    special_binning: Union[None, BinsInputType, Binning] = None,
    cut_string: Optional[str] = None,
    fig_size: Tuple[float, float] = (6, 4.5),
    fig_dpi: int = 100,
    y_lim_percentage: Optional[float] = None,
    output_name: Optional[str] = None,
    output_path: Optional[PathType] = None,
) -> Tuple[FigureType, AxesType, Dict[str, Any]]:
    if weight_column is not None:
        assert isinstance(weight_column, str)
        assert weight_column in df.columns
    if (output_name is not None) and (output_path is None):
        raise ValueError("If a output file name is provided, the output_path must also be set!")

    if sig_mc_matching_collection is None:
        sig_mc_matching_collection = default_sig_mc_matching_collection
        assert sig_mc_matching_collection.sig_id_col_name == default_sig_id_col_name, (
            sig_mc_matching_collection.sig_id_col_name,
            default_sig_id_col_name,
        )

    sig_id_col_name: str = sig_mc_matching_collection.get_sig_id_column_name_for_df(df=df)

    color_dict: Dict[str, str] = sig_mc_matching_collection.get_group_name_to_color_mapping()
    sig_id_group_name_to_latex_label_dict: Dict[
        str, str
    ] = sig_mc_matching_collection.get_group_name_to_latex_label_mapping()
    sig_id_mapping: Dict[str, Union[int, Tuple[int, ...]]] = sig_mc_matching_collection.get_sig_id_mapping_for(
        sig_id_column=sig_id_col_name
    )

    shp = StackedHistogramPlot(variable)

    for comp in sig_mc_matching_collection.plot_order:
        mc_sig_id: Union[int, Tuple[int, ...]] = sig_id_mapping[comp]

        if isinstance(mc_sig_id, int):
            mc_sig_match: np.ndarray = df[sig_id_col_name] == mc_sig_id
        elif isinstance(mc_sig_id, tuple):
            mc_sig_match = df[sig_id_col_name].isin(mc_sig_id)
        else:
            raise TypeError(f"Unexpected type for mc_sig_id: {type(mc_sig_id)} = {mc_sig_id}")

        if cut_string:
            shp.add_component(
                label=sig_id_group_name_to_latex_label_dict[comp],
                data=df[mc_sig_match].query(cut_string),
                special_binning=special_binning,
                weights=weight_column,
                color=color_dict[comp],
            )
        else:
            shp.add_component(
                label=sig_id_group_name_to_latex_label_dict[comp],
                data=df[mc_sig_match],
                special_binning=special_binning,
                weights=weight_column,
                color=color_dict[comp],
            )

    fig = plt.figure(num=None, figsize=fig_size, dpi=fig_dpi, facecolor="w", edgecolor="k")
    ax = plt.gca()

    shp.plot_on(ax=ax)

    if y_lim_percentage is not None:
        assert 0.0 < y_lim_percentage < 1.0
        y_min, y_max = ax.get_ylim()
        ax.set_ylim(bottom=y_min, top=y_lim_percentage * y_max)

    if output_name:
        export(fig, output_name, output_path)

    return fig, ax, {"filename": output_name, "target_dir": output_path}


def plot_stacked_for_all_modes(
    df: pd.DataFrame,
    variable: HistVariable,
    weight_column: Optional[str] = None,
    special_binning: Union[None, BinsInputType, Binning, Dict[str, BinsInputType]] = None,
    cut_string: Optional[str] = None,
    sig_mc_matching_collection: Optional[MCMatchingCollection] = None,
    reco_mode_info_collection: Optional[RecoModeCollection] = None,
    y_lim_percentage: Optional[float] = None,
    fig_size: Tuple[float, float] = (6, 4.5),
    fig_dpi: int = 100,
    additional_text: Optional[str] = None,
    output_name: Optional[str] = None,
    base_output_path: Optional[PathType] = None,
    close_figure: bool = False,
    dry_run: bool = False,
) -> Generator[Tuple[Optional[FigureType], Optional[AxesType], str, Optional[PathType]], None, None]:

    if (output_name is not None) and (base_output_path is None):
        raise ValueError("If a output file name is provided, the output_path must also be set!")

    if sig_mc_matching_collection is None:
        sig_mc_matching_collection = default_sig_mc_matching_collection
        assert sig_mc_matching_collection.sig_id_col_name == default_sig_id_col_name, (
            sig_mc_matching_collection.sig_id_col_name,
            default_sig_id_col_name,
        )

    color_dict: Dict[str, str] = sig_mc_matching_collection.get_group_name_to_color_mapping()
    sig_id_group_name_to_latex_label_dict: Dict[
        str, str
    ] = sig_mc_matching_collection.get_group_name_to_latex_label_mapping()

    _sig_id_col_name: Optional[str] = None
    _sig_id_mapping: Optional[Dict[str, Union[int, Tuple[int, ...]]]] = None

    if reco_mode_info_collection is None:
        reco_mode_info_collection = BSigRecoModes.main_modes

    dec_modes: RecoModeDictType = reco_mode_info_collection.main_reco_mode_infos_mapping
    dec_titles: RecoModeGroupNameToTitleDictType = reco_mode_info_collection.get_reco_mode_name_to_latex_label_mapping(
        short=False
    )

    all_mode_ids = tuple(d_id for d_ids in dec_modes.values() for d_id in d_ids)

    mode_infos = [(k, v) for k, v in dec_modes.items()]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        logger.info(f"Looking at Channel {mode_name} ({mode_ids}):")
        output_key = f"{output_name}_{mode_name}" if output_name else ""
        output_path: Optional[PathType] = (
            os.path.join(base_output_path, f"{output_key}.pdf") if base_output_path else None
        )

        if dry_run:
            yield None, None, output_key, output_path
        else:
            if weight_column is not None:
                assert isinstance(weight_column, str)
                assert weight_column in df.columns

            if isinstance(special_binning, dict):
                assert mode_name in special_binning, (mode_name, special_binning.keys())
                current_special_binning: Union[None, BinsInputType, Binning] = special_binning[mode_name]
            else:
                current_special_binning = special_binning

            if _sig_id_col_name is None:
                sig_id_col_name: str = sig_mc_matching_collection.get_sig_id_column_name_for_df(df=df)
                _sig_id_col_name = sig_id_col_name
            else:
                sig_id_col_name = _sig_id_col_name

            if _sig_id_mapping is None:
                sig_id_mapping: Dict[
                    str, Union[int, Tuple[int, ...]]
                ] = sig_mc_matching_collection.get_sig_id_mapping_for(sig_id_column=sig_id_col_name)
                _sig_id_mapping = copy.copy(sig_id_mapping)
            else:
                sig_id_mapping = _sig_id_mapping

            reco_mode_match: np.ndarray = df[Cols.decay_mode_column_name].isin(mode_ids)

            shp = StackedHistogramPlot(variable)
            for comp in sig_mc_matching_collection.plot_order:
                mc_sig_id: Union[int, Tuple[int, ...]] = sig_id_mapping[comp]

                if isinstance(mc_sig_id, int):
                    mc_sig_match: np.ndarray = df[sig_id_col_name] == mc_sig_id
                elif isinstance(mc_sig_id, tuple):
                    mc_sig_match = df[sig_id_col_name].isin(mc_sig_id)
                else:
                    raise TypeError(f"Unexpected type for mc_sig_id: {type(mc_sig_id)} = {mc_sig_id}")

                logger.info(
                    f"\t- component {comp} with sig-IDs {mc_sig_id} "
                    f"(#Cands: {np.sum(reco_mode_match & mc_sig_match)}) (sig-ID column: {sig_id_col_name})"
                )

                if cut_string:
                    shp.add_component(
                        label=sig_id_group_name_to_latex_label_dict[comp],
                        data=df[reco_mode_match & mc_sig_match].query(cut_string),
                        special_binning=current_special_binning,
                        weights=weight_column,
                        color=color_dict[comp],
                    )
                else:
                    shp.add_component(
                        label=sig_id_group_name_to_latex_label_dict[comp],
                        data=df[reco_mode_match & mc_sig_match],
                        special_binning=current_special_binning,
                        weights=weight_column,
                        color=color_dict[comp],
                    )

            fig = plt.figure(num=None, figsize=fig_size, dpi=fig_dpi, facecolor="w", edgecolor="k")
            ax = plt.gca()

            ax.set_title(dec_titles[mode_name], loc="right")

            if additional_text:
                ax.text(
                    0.05,
                    0.95,
                    additional_text,
                    horizontalalignment="left",
                    verticalalignment="top",
                    transform=ax.transAxes,
                )

            shp.plot_on(ax=ax)

            if y_lim_percentage is not None:
                assert 0.0 < y_lim_percentage < 1.0
                y_min, y_max = ax.get_ylim()
                ax.set_ylim(bottom=y_min, top=y_lim_percentage * y_max)

            if output_name:
                export(fig=fig, filename=output_key, target_dir=base_output_path, close_figure=close_figure)

            yield fig, ax, output_key, output_path


def plot_chi_squared_toys(
    toys: np.ndarray,
    bin_edges: np.ndarray,
    ndf: int,
    observed_chi2: float,
    fig_size: Tuple[Union[int, float], Union[int, float]] = (5.5, 5.0),
    full_output_path: Optional[PathType] = None,
    tag: Optional[str] = None,
) -> Tuple[FigureType, AxesType]:
    fig, ax = plt.subplots(1, 1, figsize=fig_size, dpi=200)

    ax.hist(
        toys,
        bins=bin_edges,
        color=plot_style.KITColors.kit_blue,
        label=r"Toy $\chi^2$ Distribution",
        density=True,
    )

    x_vals = np.linspace(start=bin_edges[0], stop=np.ceil(bin_edges[-1]), num=100)
    ax.plot(
        x_vals,
        chi2.pdf(x_vals, ndf),
        c=plot_style.KITColors.kit_red,
        ls="-",
        lw=2.0,
        label=r"$\chi^2$ Distribution",
    )

    ax.axvline(
        x=observed_chi2,
        c=plot_style.KITColors.kit_orange,
        ls="--",
        lw=2.0,
        label=rf"Observed $\chi^2$ = {observed_chi2:0.2f}",
    )

    ax.set_xlabel(r"$\chi^2$")
    ax.legend(frameon=False)

    if full_output_path is not None:
        output_path, base_name = os.path.split(full_output_path)
        output_name, _ = os.path.splitext(base_name)
        if tag is not None:
            output_name = f"{output_name}_{tag}"
        export(fig, output_name, output_path)

    return fig, ax
