"""
Definition of Systematics
    via collection of weight variations...

Felix Metzner 2019
"""
import numpy as np
import pandas as pd

from dataclasses import dataclass, asdict
from typing import Optional, Tuple, List, Dict, Any

from templatefitter.binned_distributions.systematics import SystematicsInputType

from rdstar.latex_labels.latex_particle_names import LatexParticleNames

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.corrections.qed_correction import new_coulomb_correction_col_name
from rdstar.offline_analysis.mc_matching.mc_matching_info import ExtendedMCMatchingIDsQueryInfo
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector as KSCorr
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector as TrkCorr
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection as PIDCorr

# from rdstar.offline_analysis.corrections.fei_tag_correction import InclusiveTagCorrection as InclTagCorr

from rdstar.mc_reweighting.BtoDlnu.BtoDlnu_weights import BtoDFFReweighting as BtoD_FF
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting as BtoDs_FF
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import BtoDststFFReweighting as BtoDss_FF
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import DststType as BtoDststFF_decay_types


__all__ = [
    "get_lepton_id_ratio_col_name_for",
    "get_lepton_id_variation_col_prefix_for",
    "split_lepton_id_cols_for",
    "SysColsInfo",
    #
    "full_systematics_name",
    "full_systematics_cols",
    "full_systematics_dict",
    "multiplicative_eff_sys_info_names",
    #
    "new_full_systematics_cols",
    "new_full_systematics_dict",
    "new_shape_sys_info_names",
    #
    "get_systematics",
]


def get_lepton_id_ratio_col_name_for(base_name: Optional[str]) -> str:
    if base_name is None:
        base_name = "lepton"
        ratio_col_name: str = f"__weight_pid_corr_{base_name}_ratio__"
        assert ratio_col_name == PIDCorr.pid_l_corr_col_name, (ratio_col_name, PIDCorr.pid_l_corr_col_name)
        return ratio_col_name
    else:
        assert base_name != "lepton", base_name
        return f"__weight_pid_corr_{base_name}_ratio__"


def get_lepton_id_variation_col_prefix_for(base_name: Optional[str]) -> str:
    if base_name is None:
        base_name = "lepton"
        var_col_prefix: str = f"__weight_pid_corr_{base_name}_variation_"
        assert var_col_prefix == PIDCorr.pid_l_corr_var_prefix, (var_col_prefix, PIDCorr.pid_l_corr_col_name)
        return var_col_prefix
    else:
        assert base_name != "lepton", base_name
        return f"__weight_pid_corr_{base_name}_variation_"


new_lepton_id_corr_col_name: str = "__weight_pid_corr_lepton_id_ratio__"
new_lepton_id_corr_var_col_prefix: str = "__weight_pid_corr_lepton_id_variation_"
new_lepton_fake_id_corr_col_name: str = "__weight_pid_corr_lepton_fake_id_ratio__"
new_lepton_fake_id_corr_var_col_prefix: str = "__weight_pid_corr_lepton_fake_id_variation_"


def split_lepton_id_cols_for(
    df: pd.DataFrame,
    base_lepton_id_name: Optional[str] = None,
) -> None:

    old_lepton_id_col_name: str = get_lepton_id_ratio_col_name_for(base_name=base_lepton_id_name)
    old_lepton_id_var_col_prefix: str = get_lepton_id_variation_col_prefix_for(base_name=base_lepton_id_name)

    if base_lepton_id_name is None:
        assert old_lepton_id_col_name == PIDCorr.pid_l_corr_col_name, old_lepton_id_col_name
        assert old_lepton_id_var_col_prefix == PIDCorr.pid_l_corr_var_prefix, old_lepton_id_var_col_prefix

    assert old_lepton_id_col_name in df.columns, old_lepton_id_col_name
    assert len([c for c in df.columns if c.startswith(old_lepton_id_var_col_prefix)]) > 0, old_lepton_id_var_col_prefix

    old_l_id_var_columns: List[str] = [c for c in df.columns if c.startswith(old_lepton_id_var_col_prefix)]
    assert len(old_l_id_var_columns) == 20, (
        old_lepton_id_var_col_prefix,
        len(old_l_id_var_columns),
        old_l_id_var_columns,
    )

    assert new_lepton_id_corr_col_name not in df.columns, new_lepton_id_corr_col_name
    assert len([c for c in df.columns if c.startswith(new_lepton_id_corr_var_col_prefix)]) == 0, (
        new_lepton_id_corr_col_name,
        [c for c in df.columns if c.startswith(new_lepton_id_corr_var_col_prefix)],
    )
    assert new_lepton_fake_id_corr_col_name not in df.columns, new_lepton_fake_id_corr_col_name
    assert len([c for c in df.columns if c.startswith(new_lepton_fake_id_corr_var_col_prefix)]) == 0, (
        new_lepton_fake_id_corr_col_name,
        [c for c in df.columns if c.startswith(new_lepton_fake_id_corr_var_col_prefix)],
    )

    assert ExtendedMCMatchingIDsQueryInfo.sig_col_name in df.columns, ExtendedMCMatchingIDsQueryInfo.sig_col_name

    lepton_fake_sig_ids: Tuple[int, ...] = (
        ExtendedMCMatchingIDsQueryInfo.LeptonFake_HBKG.id,
        ExtendedMCMatchingIDsQueryInfo.LeptonRecoFake_HBKG.id,
    )
    fake_lepton_mask: pd.Series = df[ExtendedMCMatchingIDsQueryInfo.sig_col_name].isin(lepton_fake_sig_ids)

    df.loc[:, new_lepton_id_corr_col_name] = df[old_lepton_id_col_name].values
    df.loc[:, new_lepton_fake_id_corr_col_name] = df[old_lepton_id_col_name].values

    df.loc[fake_lepton_mask, new_lepton_id_corr_col_name] = 1.0
    df.loc[~fake_lepton_mask, new_lepton_fake_id_corr_col_name] = 1.0

    for _var_col in old_l_id_var_columns:
        _l_id_var_col_name: str = _var_col.replace(old_lepton_id_var_col_prefix, new_lepton_id_corr_var_col_prefix)
        assert _l_id_var_col_name.startswith(new_lepton_id_corr_var_col_prefix), (
            _l_id_var_col_name,
            new_lepton_id_corr_var_col_prefix,
        )
        _lf_id_var_col_name: str = _var_col.replace(old_lepton_id_var_col_prefix, new_lepton_fake_id_corr_var_col_prefix)
        assert _lf_id_var_col_name.startswith(new_lepton_fake_id_corr_var_col_prefix), (
            _lf_id_var_col_name,
            new_lepton_fake_id_corr_var_col_prefix,
        )

        df.loc[:, _l_id_var_col_name] = df[_var_col].values
        df.loc[:, _lf_id_var_col_name] = df[_var_col].values

        df.loc[fake_lepton_mask, _l_id_var_col_name] = 1.0
        df.loc[~fake_lepton_mask, _lf_id_var_col_name] = 1.0


@dataclass(frozen=True)
class SysColsInfo:
    var: str
    ratio: Optional[str] = None
    from_df: bool = False
    up_down_from_df: Optional[Tuple[str, str]] = None
    ends_with_up_down_tag: Optional[bool] = None
    consider_in_new_approach: bool = False
    _name: Optional[str] = None
    _latex_name: Optional[str] = None

    def __post_init__(self) -> None:
        if self.from_df and self.ratio is None:
            raise ValueError("Field 'from_df' cannot be set to True if 'ratio' is None!")

        if self._name is not None:
            assert self._name
        if self._latex_name is not None:
            assert self._latex_name

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name
        else:
            return self.var

    @property
    def latex_name(self) -> str:
        if self._latex_name is not None:
            return self._latex_name
        else:
            return self.name

    @property
    def sys_info_type(self) -> str:
        if self.ratio is None:
            return "uncertainty"
        else:
            if self.from_df is False:
                return "up_down_uncertainty"
            else:
                return "variations"

    def get_systematics_input_for_df(self, df: pd.DataFrame) -> SystematicsInputType:
        if self.ratio is None:
            assert self.var in df.columns, self.var
            assert not pd.isnull(df[self.var]).any(), (self.var, df.loc[pd.isnull(df[self.var]), Cols.data_type])
            return None, self.var
        else:
            assert self.ratio in df.columns, self.ratio
            if self.from_df:
                if self.up_down_from_df is None:
                    var_cols: List[str] = [c for c in df.columns if c.startswith(self.var)]
                    assert len(var_cols) > 0, (self.ratio, self.var)
                    relevant_cols: List[str] = [self.ratio, *var_cols]
                    assert all(not pd.isnull(df[c]).any() for c in relevant_cols), (
                        relevant_cols,
                        [(c, df.loc[pd.isnull(c), Cols.data_type]) for c in relevant_cols],
                    )
                    return self.ratio, var_cols
                elif isinstance(self.up_down_from_df, tuple) and len(self.up_down_from_df) == 2:
                    assert all(isinstance(s, str) for s in self.up_down_from_df), (
                        self.up_down_from_df,
                        [type(c) for c in self.up_down_from_df],
                    )
                    assert self.ends_with_up_down_tag is not None

                    up_col_str, down_col_str = self.up_down_from_df  # type: str, str

                    if self.ends_with_up_down_tag is False:
                        up_var_cols: List[str] = [c for c in df.columns if c.startswith(self.var + up_col_str)]
                    else:
                        up_var_cols = [c for c in df.columns if c.startswith(self.var) and c.endswith(up_col_str)]

                    down_var_cols: List[str] = [c.replace(up_col_str, down_col_str) for c in up_var_cols]

                    assert all(c in df.columns for c in up_var_cols), [c for c in up_var_cols if c not in df.columns]
                    assert all(c in df.columns for c in down_var_cols), [c for c in down_var_cols if c not in df.columns]

                    assert len(up_var_cols) > 0, (self.ratio, self.var, up_var_cols)
                    assert len(down_var_cols) > 0, (self.ratio, self.var, down_var_cols)

                    relevant_cols = [self.ratio, *up_var_cols, *down_var_cols]
                    assert not any(np.any(pd.isnull(df[c])) for c in relevant_cols), (
                        relevant_cols,
                        [(c, df.loc[pd.isnull(c), Cols.data_type]) for c in relevant_cols],
                    )
                    return self.ratio, [*up_var_cols, *down_var_cols]
                else:
                    raise RuntimeError(f"The case 'up_down_from_df = {self.up_down_from_df}' is not covered!")
            else:
                assert self.var in df.columns, (self.ratio, self.var)
                assert all(not pd.isnull(df[c]).any() for c in (self.ratio, self.var)), (
                    self.ratio,
                    self.var,
                    [(c, df.loc[pd.isnull(c), Cols.data_type]) for c in (self.ratio, self.var)],
                )
                return self.ratio, self.var

    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(input_dict) -> "SysColsInfo":
        return SysColsInfo(**input_dict)


full_systematics_name: str = "full_sys"

_full_systematics_cols: Tuple[SysColsInfo, ...] = (
    SysColsInfo(
        var=PIDCorr.pid_l_corr_var_prefix,
        ratio=PIDCorr.pid_l_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="lepton_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.lepton_str[1:] + " ID",
    ),
    SysColsInfo(
        var=PIDCorr.pid_pi_corr_var_prefix,
        ratio=PIDCorr.pid_pi_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="pion_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.pi_str[1:] + " ID",
    ),
    SysColsInfo(
        var=PIDCorr.pid_piFromK_corr_var_prefix,
        ratio=PIDCorr.pid_piFromK_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="pion_fake_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.pi_str[1:] + " Fake ID",
    ),
    SysColsInfo(
        var=PIDCorr.pid_k_corr_var_prefix,
        ratio=PIDCorr.pid_k_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="kaon_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.k_str[1:] + " ID",
    ),
    SysColsInfo(
        var=PIDCorr.pid_kFromPi_corr_var_prefix,
        ratio=PIDCorr.pid_kFromPi_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="kaon_fake_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.k_str[1:] + " Fake ID",
    ),
    SysColsInfo(
        var=TrkCorr.slow_pi_corr_var_prefix,
        ratio=TrkCorr.slow_pi_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="slow_pion",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.pi_slow_str[1:],
    ),
    SysColsInfo(
        var=KSCorr.k_short_corr_var_prefix,
        ratio=KSCorr.k_short_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="k_short",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.k_short_str[1:],
    ),
    SysColsInfo(
        var=TrkCorr.tracking_systematic_name,
        ratio=None,
        from_df=False,
        consider_in_new_approach=True,
        _name="tracking_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\,$ Tracking",
    ),
    SysColsInfo(
        var=TagCorrection.tagging_corr_uncert_name,
        ratio=TagCorrection.tagging_correction_name,
        from_df=False,
        consider_in_new_approach=False,
        _name="tagging_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.b_tag_str[1:],
    ),
    # SysColsInfo(
    #     var=InclTagCorr.tagging_corr_uncert_prefix,
    #     ratio=InclTagCorr.tagging_correction_name,
    #     from_df=True,
    #     _name="tagging_sys",
    #     _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.b_tag_str[1:],
    # ),
    # FF Systematics (Cov Matrix is calculated differently than Max did it!):
    SysColsInfo(
        var=BtoD_FF.base_var_weight_name,
        ratio=BtoD_FF.nom_weight_name,
        from_df=True,
        up_down_from_df=("_up", "_down"),
        ends_with_up_down_tag=True,
        consider_in_new_approach=True,
        _name="d_ff_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_str[1:] + " FF",
    ),
    SysColsInfo(
        var=BtoDs_FF.base_var_weight_name,
        ratio=BtoDs_FF.nom_weight_name,
        from_df=True,
        up_down_from_df=("_up", "_down"),
        ends_with_up_down_tag=True,
        consider_in_new_approach=True,
        _name="d_st_ff_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_st_str[1:] + " FF",
    ),
)


d_st_st_names = {
    BtoDststFF_decay_types.DONE: "d_one_d_st_st_ff_sys",
    BtoDststFF_decay_types.DTWOSTAR: "d_twostar_d_st_st_ff_sys",
    BtoDststFF_decay_types.DPRIMEONE: "d_primeone_d_st_st_ff_sys",
    BtoDststFF_decay_types.DZEROSTAR: "d_zerostar_d_st_st_ff_sys",
}


d_st_st_latex_names = {
    BtoDststFF_decay_types.DONE: r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_one_str[1:] + " FF",
    BtoDststFF_decay_types.DTWOSTAR: r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_two_st_str[1:] + " FF",
    BtoDststFF_decay_types.DPRIMEONE: r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_prime_one_str[1:] + " FF",
    BtoDststFF_decay_types.DZEROSTAR: r"$\sigma^\mathrm{sys}\," + LatexParticleNames.d_sszero_st_str[1:] + " FF",
}


_dss_systematics_cols_infos: Tuple[SysColsInfo, ...] = tuple(
    SysColsInfo(
        var=BtoDss_FF.get_var_dstst_weight_name(dstst_type=dstst_decay_type.value),
        ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type=dstst_decay_type.value),
        from_df=True,
        up_down_from_df=("_up", "_down"),
        ends_with_up_down_tag=True,
        consider_in_new_approach=True,
        _name=d_st_st_names[dstst_decay_type],
        _latex_name=d_st_st_latex_names[dstst_decay_type],
    )
    for dstst_decay_type in BtoDststFF_decay_types
)


full_systematics_cols: Tuple[SysColsInfo, ...] = tuple(list(_full_systematics_cols) + list(_dss_systematics_cols_infos))


assert all(v.name is not None and v.name for v in full_systematics_cols), [
    v.var for v in full_systematics_cols if v.name is None or v.name == ""
]
full_systematics_dict: Dict[str, SysColsInfo] = {v.name: v for v in full_systematics_cols}


multiplicative_eff_sys_info_names: Tuple[str, ...] = (
    "d_ff_sys",
    "d_st_ff_sys",
    "lepton_id",
    "slow_pion",
    "pion_id",
    "pion_fake_id",
    "kaon_id",
    "kaon_fake_id",
    "k_short",
)
_sys_info_names_for_eff: List[str] = [s.name for s in full_systematics_cols if s.sys_info_type == "variations"]
assert all(sn in _sys_info_names_for_eff for sn in multiplicative_eff_sys_info_names), [
    sn for sn in multiplicative_eff_sys_info_names if sn not in _sys_info_names_for_eff
]

_new_multiplicative_systematics_entries: Tuple[SysColsInfo, ...] = (
    SysColsInfo(
        var="HAMMER_weight_Norm_",
        ratio="HAMMER_weight_Norm",
        from_df=True,
        up_down_from_df=("up", "down"),
        ends_with_up_down_tag=False,
        consider_in_new_approach=True,
        _name="hammer_norm_ff_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + " Norm FF",
    ),
    SysColsInfo(
        var="HAMMER_weight_Signal_",
        ratio="HAMMER_weight_Signal",
        from_df=True,
        up_down_from_df=("up", "down"),
        ends_with_up_down_tag=False,
        consider_in_new_approach=True,
        _name="hammer_sig_ff_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + " Signal FF",
    ),
    SysColsInfo(
        var=new_lepton_id_corr_var_col_prefix,
        ratio=new_lepton_id_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="new_lepton_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.lepton_str[1:] + " ID",
    ),
    SysColsInfo(
        var=new_lepton_fake_id_corr_var_col_prefix,
        ratio=new_lepton_fake_id_corr_col_name,
        from_df=True,
        consider_in_new_approach=True,
        _name="new_lepton_fake_id",
        _latex_name=r"$\sigma^\mathrm{sys}\," + LatexParticleNames.lepton_str[1:] + " Fake ID",
    ),
    SysColsInfo(
        var=new_coulomb_correction_col_name,
        ratio=None,
        from_df=False,
        consider_in_new_approach=True,
        _name="qed_shape_correction_sys",
        _latex_name=r"$\sigma^\mathrm{sys}\," + " QED Corr.",
    ),
)

new_full_systematics_dict: Dict[str, SysColsInfo] = {v.name: v for v in full_systematics_cols}
new_full_systematics_dict.update({v.name: v for v in _new_multiplicative_systematics_entries})

new_full_systematics_cols: Tuple[SysColsInfo, ...] = tuple(v for v in new_full_systematics_dict.values())


new_shape_sys_info_names: Tuple[str, ...] = (
    "hammer_norm_ff_sys",
    "hammer_sig_ff_sys",
    #
    "d_ff_sys",
    "d_st_ff_sys",
    #
    "lepton_id",
    "new_lepton_id",
    "new_lepton_fake_id",
    #
    "slow_pion",
    #
    "pion_id",
    "pion_fake_id",
    "kaon_id",
    "kaon_fake_id",
    #
    "k_short",
    #
    "d_one_d_st_st_ff_sys",
    "d_twostar_d_st_st_ff_sys",
    "d_primeone_d_st_st_ff_sys",
    "d_zerostar_d_st_st_ff_sys",
    #
    "tracking_sys",
    #
    "qed_shape_correction_sys",
)


_new_sys_info_names_for_shape: List[str] = [s.name for s in new_full_systematics_cols if s.consider_in_new_approach]
assert all(sn in _new_sys_info_names_for_shape for sn in new_shape_sys_info_names), [
    sn for sn in new_shape_sys_info_names if sn not in _new_sys_info_names_for_shape
]


def get_systematics(df: pd.DataFrame, sys_cols_infos: Tuple[SysColsInfo, ...]) -> SystematicsInputType:
    if not sys_cols_infos or sys_cols_infos is None:
        return None

    assert all(isinstance(info, SysColsInfo) for info in sys_cols_infos), [type(info) for info in sys_cols_infos]

    return [sci.get_systematics_input_for_df(df=df) for sci in sys_cols_infos]
