"""
Tasks performing a evaluation of
    - the impact of cuts on a given variable on efficiencies for data and MC in different reconstruction modes
    - general differences in number of events in data and MC for different reconstruction modes

Felix Metzner 2021
"""

import os
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits

from typing import Optional, Tuple, Dict, Generator, ClassVar

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.validation.efficiency_validation import EfficiencyDependencyEvaluator, EffDepEvalInfo

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.data_handling.selected_data_handling import DataSampleLoader
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import ResolutionEvaluationManagerTask


__all__ = [
    "EfficiencyValidationTask",
    "EfficiencyValidationWrapperTask",
]


@inherits(ResolutionEvaluationManagerTask)
class EfficiencyValidationWrapperTask(luigi.WrapperTask):

    relevant_selections: Tuple[str, ...] = SelectionSetups.rdstar_selections

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert self.selection_name in self.relevant_selections, (self.selection_name, self.relevant_selections)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for efficiency_evaluation_setup_key, eval_info in EfficiencyDependencyEvaluator.setup_dict.items():
            for sample_setup_keys, side_band_keys in eval_info.sample_setup_keys_to_sideband_keys_map.items():
                for sample_setup_key in sample_setup_keys:
                    for side_band_key in side_band_keys:
                        for use_inclusive_fei_calibration in (True, False):
                            yield self.clone(
                                EfficiencyValidationTask,
                                efficiency_evaluation_setup_key=efficiency_evaluation_setup_key,
                                sample_setup_key=sample_setup_key,
                                side_band_key=side_band_key,
                                use_inclusive_fei_calibration=use_inclusive_fei_calibration,
                            )


@inherits(EfficiencyValidationWrapperTask)
class EfficiencyValidationTask(utilities.BaseOfflineTask):
    efficiency_evaluation_setup_key: str = luigi.Parameter(positional=False)
    sample_setup_key: str = luigi.Parameter(positional=False)
    side_band_key: str = luigi.Parameter(positional=False)
    use_inclusive_fei_calibration: bool = luigi.BoolParameter(positional=False)

    default_weight_dir_name: ClassVar[str] = "default_weight_definition"
    corr_weight_dir_name: ClassVar[str] = "weights_with_inclusive_fei_corr"

    correction_providing_setup_name: ClassVar[str] = EfficiencyDependencyEvaluator.correction_providing_setup_name

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path: PathType = os.path.join(
            self.working_dir,
            EfficiencyDependencyEvaluator.base_dir_name,
            self.selection_name,
            self.sample_setup_key,
            self.side_band_key,
            self.corr_weight_dir_name if self.use_inclusive_fei_calibration else self.default_weight_dir_name,
        )

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.data_sample_key: str = SampleSetupHandler.get_data_sample_key(sample_name=self.sample_setup_key)
        self.mc_sample_keys: Tuple[str, ...] = SampleSetupHandler.get_mc_sample_keys(sample_name=self.sample_setup_key)

        self._input_dict: Optional[Dict[str, PathType]] = None
        self._data_loader: Optional[DataSampleLoader] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def data_loader(self) -> DataSampleLoader:
        if self._data_loader is None:
            sideband_identifiers = DataSampleLoader.get_identifiers_from_sideband_key(sideband_key=self.side_band_key)
            self._data_loader = DataSampleLoader(
                input_dict=self.input_dict,
                sample_setup_key=self.sample_setup_key,
                cs_tag=self.cs_tag,
                selection_name=self.selection_name,
                input_level=2,
                sideband_identifiers=sideband_identifiers,
            )

        return self._data_loader

    @staticmethod
    def get_correction_providing_setup_info() -> EffDepEvalInfo:
        return EfficiencyDependencyEvaluator.setup_dict[EfficiencyValidationTask.correction_providing_setup_name]

    def do_efficiency_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        eff_eval_setup: EffDepEvalInfo = EfficiencyDependencyEvaluator.setup_dict[self.efficiency_evaluation_setup_key]

        validator: EfficiencyDependencyEvaluator = EfficiencyDependencyEvaluator(
            evaluation_hist_vars=eff_eval_setup.evaluation_hist_vars,
            cut_information=eff_eval_setup.cut_information,
            reco_mode_info_collection=eff_eval_setup.reco_mode_info_collection,
            weight_column_name=eff_eval_setup.weight_column_name,
            result_dir_path=self.base_path,
        )

        data_df, mc_df = (None, None)  # type: Optional[pd.DataFrame], Optional[pd.DataFrame]
        if not dry_run:
            data_df, mc_df = self.data_loader.load_data_and_mc()

            if self.use_inclusive_fei_calibration:
                WeightManager.switch_fei_calibration(
                    df=mc_df,
                    new_fei_calib_col=WeightManager.used_fei_calib_col,
                    main_weight_col_name=eff_eval_setup.weight_column_name,
                )

        return validator.perform_efficiency_evaluation(
            mc_sample=mc_df,
            data_sample=data_df,
            dry_run=dry_run,
        )

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_efficiency_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_efficiency_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )
