"""
Task to evaluate a possible shift in the squared missing mass distributions.

Felix Metzner 2021
"""

import os
import json
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits
from typing import Optional, List, Tuple, Dict, Generator, Any

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MM2ShiftFitter
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import ResolutionEvaluationManagerTask
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit_setups import (
    ShiftEvaluationInfo,
    mm2_shift_eval_infos_tuple,
)


__all__ = [
    "MM2ShiftEvaluationTask",
]


@inherits(ResolutionEvaluationManagerTask)
class MM2ShiftEvaluationTask(luigi.WrapperTask):
    shift_eval_infos_dict = {v.name: v for v in mm2_shift_eval_infos_tuple}  # type: Dict[str, ShiftEvaluationInfo]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for sample_setup_key in SampleSetupHandler.on_resonance_sample_setup_names:
            for eval_info_key_definition_key in self.shift_eval_infos_dict.keys():
                for minimizer_method in MM2ShiftFitter.minimizer_methods:
                    for chi2_method in MM2ShiftFitter.chi2_methods:
                        yield self.clone(
                            MM2ShiftFitTask,
                            sample_setup_key=sample_setup_key,
                            eval_info_key=eval_info_key_definition_key,
                            minimizer_method=minimizer_method,
                            chi2_calc_method=chi2_method,
                        )


@inherits(MM2ShiftEvaluationTask)
class MM2ShiftFitTask(utilities.BaseOfflineTask):
    sample_setup_key = luigi.Parameter(positional=False)  # type: str
    eval_info_key = luigi.Parameter(positional=False)  # type: str
    minimizer_method = luigi.Parameter(positional=False)  # type: str
    chi2_calc_method = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "MM2ShiftFitter"  # type: str
    eval_result_file_name = "mm2_shift_fit_result"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.selection_name,
            self.sample_setup_key,
            self.eval_info_key,
            self.chi2_calc_method,
            self.minimizer_method,
        )  # type: PathType

        self.eval_info = MM2ShiftEvaluationTask.shift_eval_infos_dict[self.eval_info_key]  # type: ShiftEvaluationInfo

        self.data_sample_key = SampleSetupHandler.get_data_sample_key(sample_name=self.sample_setup_key)  # type: str
        self.mc_sample_keys = SampleSetupHandler.get_mc_sample_keys(
            sample_name=self.sample_setup_key
        )  # type: Tuple[str, ...]

        self.sideband_key = "q2_sideband"  # type: str

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def mm2_hist_var(self) -> HistVariable:
        return self.eval_info.hist_var

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_mc_and_data_file_paths(self) -> Optional[Tuple[List[PathType], PathType]]:
        key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"  # type: str
        input_appendix = f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"  # type: str

        full_key = f"{key_prefix}{self.sideband_key}"  # type: str

        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (k.startswith(f"{mc_key}_{input_appendix}_") and k.endswith(full_key) and os.path.splitext(p)[1] == ".h5")
        ]  # type: List[PathType]

        if len(mc_paths) == 0:
            return None

        assert len(mc_paths) == len(self.mc_sample_keys)

        data_file_start = f"{self.data_sample_key}_{input_appendix}_"  # type: str
        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (k.startswith(data_file_start) and k.endswith(full_key) and os.path.splitext(p)[1] == ".h5")
        ]  # type: List[PathType]
        assert len(data_paths) == 1, data_paths

        return mc_paths, data_paths[0]

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        mc_and_data_paths = self.get_mc_and_data_file_paths()
        assert mc_and_data_paths is not None
        mc_paths, data_path = mc_and_data_paths  # type: List[PathType], PathType

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame)

        if self.eval_info.fei_weight_col is not None:
            WeightManager.switch_fei_calibration(
                df=mc_sample,
                main_weight_col_name=WeightManager.final_weight_with_additions_col,
                new_fei_calib_col=self.eval_info.fei_weight_col,
            )

        return mc_sample, data_sample

    def do_mm2_shift_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_file_path = os.path.join(self.base_path, f"{self.eval_result_file_name}.json")  # type: PathType
        tmp_output_file_path = os.path.join(self.base_path, f"{self.eval_result_file_name}.json")  # type: PathType
        output_dict = {self.eval_result_file_name: output_file_path}  # type: Dict[str, PathType]

        if dry_run:
            return output_dict

        mc_df, data_df = self.get_mc_and_data_sample()  # type: pd.DataFrame, pd.DataFrame
        fitter = MM2ShiftFitter(
            mc_df=mc_df,
            data_df=data_df,
            weight_column_name=WeightManager.final_weight_with_additions_col,
            mm2_hist_var=self.mm2_hist_var,
            minimizer_method=self.minimizer_method,
            chi2_method=self.chi2_calc_method,
        )

        shift_fit_result_dict = fitter.find_shift()  # type: Dict[str, Any]

        mm2_shift_result_dict = {
            "eval_info_key": self.eval_info_key,
            "sideband_key": self.sideband_key,
        }
        mm2_shift_result_dict.update(shift_fit_result_dict)

        with open(file=tmp_output_file_path, mode="w") as f:
            json.dump(mm2_shift_result_dict, fp=f, indent=2)
        os.rename(src=tmp_output_file_path, dst=output_file_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_mm2_shift_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_mm2_shift_evaluation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )
