"""
Utility Functions for Analysis Tasks

Felix Metzner 2019
"""

import os
import json
import yaml
import rdstar
import socket
import getpass

import numpy as np
import pandas as pd
import b2luigi as luigi

from functools import lru_cache
from collections.abc import Sized
from typing import Any, Union, Optional, Tuple, List, Dict, NamedTuple

from rdstar.utilities import PathType


__all__ = [
    "TaskType",
    "InputDictType",
    "ConfigCombinationsType",
    "offline_memory_resource_name",
    "cache_dir_mapping",
    "GlobalParameters",
    "BaseOfflineTask",
    "get_config_directory",
    "get_config_dir_path_for",
    "get_user_name",
    "get_host_name",
    "host_check",
    "set_numexpr_max_threads",
    "get_workers_for_host",
    "load_dict_from_json",
    "to_float_if_float",
    "fix_json_dict",
    "get_d_st_ff_precomputed_gammas_file_path",
    "decompose_event_id",
    "recreate_event_id",
    "UniqueEventIDColumnInfos",
    "add_unique_candidate_id",
]

TaskType = luigi.Task
InputDictType = Dict[str, PathType]
ConfigCombinationsType = List[Tuple[str, List[str], List[Dict[str, Any]]]]

offline_memory_resource_name = "offline_memory"  # type: str


_host_list: List[str] = [
    "deepthought",
    "portal1.etp.kit.edu",
    "bms1",
    "bms2",
    "bms3",
    "bms1-centos7",
    "bms2-centos7",
    "bms3-centos7",
]

_fake_cache_base_dir_path: str = "/ceph/fmetzner/fake_cache_dirs/"

cache_dir_mapping: Dict[str, PathType] = {
    "deepthought": "/local/scratch/ssd1/fmetzner/analysis_cache",
    "portal1.etp.kit.edu": "/local/scratch/user_space/fmetzner/analysis_cache",
}

for host in _host_list:
    if host not in cache_dir_mapping:
        cache_dir_mapping[host] = str(os.path.join(_fake_cache_base_dir_path, host.replace(".", "_"), "analysis_cache"))


class GlobalParameters(luigi.Config):
    work_host_list = luigi.ListParameter(default=[])
    copy_host_list = luigi.ListParameter(default=[])

    n_subsets_mixed = luigi.IntParameter(default=1)
    n_subsets_generic = luigi.IntParameter(default=1)


class BaseOfflineTask(luigi.Task):
    # Luigi default priority value
    default_priority = 0  # type: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def priority(self) -> int:
        try:
            if self.resources[offline_memory_resource_name] > 1:
                return int(10 * self.resources[offline_memory_resource_name])
            else:
                return 10
        except AttributeError:
            return self.default_priority
        except TypeError:
            return self.default_priority
        except KeyError:
            return self.default_priority

    def get_task_file_dir(self) -> PathType:
        assert self.working_dir is not None
        return self.working_dir

    def get_log_file_dir(self) -> PathType:
        assert self.get_task_file_dir() is not None
        return self.get_task_file_dir()

    @staticmethod
    def get_temp_file_path(
        path: PathType,
        prefix: Optional[str] = None,
    ) -> PathType:
        if prefix is None:
            prefix = "temp"
        dir_path, file_name = os.path.split(path)
        return os.path.join(dir_path, f"{prefix}_{file_name}")


def get_config_directory() -> PathType:
    pkg_dir = os.path.dirname(rdstar.__file__)  # type: PathType
    return os.path.join(pkg_dir, "configs")


def get_config_dir_path_for(production_tag: str) -> PathType:
    bd = get_config_directory()  # type: PathType
    base_production_tag, stream_str = production_tag.split("_stream", 1)  # type: str, str

    _production_mapping = {
        "prod_02ndMay2021": os.path.join(bd, f"prod_02ndMay2021_stream{stream_str}"),
        "prod_24thMarch2022": os.path.join(bd, f"prod_24thMarch2022_stream{stream_str}"),
        "prod_28thMarch2022_pilnu_reco": os.path.join(bd, "xulnu_reco", f"prod_28thMarch2022_stream{stream_str}"),
    }  # type: Dict[str, PathType]

    assert base_production_tag in _production_mapping, (production_tag, base_production_tag, _production_mapping)
    return _production_mapping[base_production_tag]


@lru_cache()
def load_yaml_config(yaml_config_path: PathType) -> Dict[Any, Any]:
    with open(yaml_config_path) as f:
        offline_config = yaml.load(f, Loader=yaml.FullLoader)
    return offline_config


def get_user_name() -> str:
    return getpass.getuser()


def get_host_name() -> str:
    return socket.gethostname()


def host_check(hosts: List[str], message: Optional[str] = None) -> bool:
    if not hosts:
        return True

    if message is None:
        message = "The task must be run on a suitable machine...\n"
        message += f"Switch to any of the following hosts:\n {hosts}"

    if not any(host in get_host_name() for host in hosts):
        raise RuntimeError(message)

    return True


def host_is_in(host_list: List[str]) -> bool:
    this_host_name = get_host_name()  # type: str
    return any(host == this_host_name for host in host_list)


def set_numexpr_max_threads() -> Tuple[int, int]:
    n_cpus_available = len(os.sched_getaffinity(0))  # type: int
    if host_is_in(host_list=["deepthought", "portal1.etp.kit.edu"]):
        n_cpus_used = 20  # type: int
    elif host_is_in(host_list=["bms1", "bms2", "bms3", "bms1-centos7", "bms2-centos7", "bms3-centos7"]):
        n_cpus_used = 20
    elif host_is_in(host_list=["ekplx71"]):
        n_cpus_used = 6
    else:
        n_cpus_used = int(n_cpus_available / 2)

    assert n_cpus_used < n_cpus_available, (n_cpus_used, n_cpus_available, get_host_name())
    os.environ["NUMEXPR_MAX_THREADS"] = f"{n_cpus_used}"
    return n_cpus_used, n_cpus_available


def get_workers_for_host() -> int:
    workers_by_host_dict: Dict[str, int] = {
        "ekplx71": 2,
        "deepthought": 8,
        "portal1.etp.kit.edu": 10,
        "bms1": 8,
        "bms2": 2,
        "bms3": 2,
        "bms1-centos7": 10,
        "bms2-centos7": 10,
        "bms3-centos7": 10,
    }
    host_name = get_host_name()
    if host_name not in workers_by_host_dict:
        raise KeyError(
            f"No instructions available for host {host_name}! "
            f"Supported are the hosts {list(workers_by_host_dict.keys())}."
        )
    return workers_by_host_dict[host_name]


def load_dict_from_json(path: PathType) -> Dict[Any, Any]:
    with open(path, "r") as in_file:
        eval_dict = json.load(in_file)
    return fix_json_dict(json_dict=eval_dict)


def to_float_if_float(input_string: str) -> Union[float, str]:
    try:
        num = float(input_string)
    except ValueError:
        return input_string
    return num


def fix_json_dict(json_dict: Union[Dict[str, Any], Any]) -> Union[Dict[Union[float, str], Any], Any]:
    if not isinstance(json_dict, dict):
        return json_dict
    reformatted_dict = {
        to_float_if_float(k): fix_json_dict(v) for k, v in json_dict.items()
    }  # type: Dict[Union[float, str], Any]
    return reformatted_dict


def get_d_st_ff_precomputed_gammas_file_path(working_dir: PathType) -> PathType:
    return os.path.join(os.path.dirname(working_dir), "dst_reweighting_gamma_integrals.json")


class UniqueEventIDColumnInfos(NamedTuple):
    event_id_col: str = "unique_event_id"
    data_type_col: str = "temp_data_type_info"
    orig_exp_col: str = "expNum"
    orig_run_col: str = "runNum"
    orig_evt_col: str = "evtNum"
    orig_other_col: str = "totalEnergyOfParticlesInList__bogamma__clfsp__bc"
    exp_col: str = "temp_exp_info"
    run_col: str = "temp_run_info"
    evt_col: str = "temp_evt_info"
    other_col: str = "temp_gamma_energy_info"
    split_str: str = "_"
    split_n: int = 4
    round_to: int = 6

    @staticmethod
    def get_unique_event_id_col_name() -> str:
        return UniqueEventIDColumnInfos().event_id_col


def add_unique_event_id(
    df: pd.DataFrame,
    data_type: str,
) -> None:
    num_cols = len(df.columns)
    col_info = UniqueEventIDColumnInfos()
    assert col_info.event_id_col not in df.columns, col_info.event_id_col

    assert col_info.orig_exp_col in df.columns, col_info.orig_exp_col
    assert col_info.orig_run_col in df.columns, col_info.orig_run_col
    assert col_info.orig_evt_col in df.columns, col_info.orig_evt_col
    assert col_info.orig_other_col in df.columns, col_info.orig_other_col

    assert col_info.data_type_col not in df.columns, col_info.data_type_col
    df.loc[:, col_info.data_type_col] = data_type

    temp_cols = [col_info.data_type_col]
    info_cols = [col_info.data_type_col]
    int_info_cols = [col_info.orig_exp_col, col_info.orig_run_col, col_info.orig_evt_col]
    float_info_cols = [col_info.orig_other_col]

    for col in int_info_cols:
        converted_col_name = f"temp_{col}"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, converted_col_name] = df[col].astype(float).astype(str)
        temp_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    for col in float_info_cols:
        rounded_col_name = f"temp_{col}_rounded"
        assert rounded_col_name not in df.columns, rounded_col_name
        converted_col_name = f"temp_{col}_str"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, rounded_col_name] = df[col].round(decimals=col_info.round_to)
        temp_cols.append(rounded_col_name)
        df.loc[:, converted_col_name] = df[rounded_col_name].astype(str)
        temp_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    assert all((not pd.isnull(df[c]).any()) for c in info_cols), [c for c in info_cols if pd.isnull(df[c]).any()]

    len_before = len(df.index)
    id_values = pd.Series(df.loc[:, info_cols].values.tolist()).str.join(col_info.split_str)  # type: pd.Series

    assert len(df.index) == len(id_values.index), (len(df.index), len(id_values.index))
    assert isinstance(df.index, Sized)
    assert (df.index == id_values.index).all()

    df.loc[:, col_info.event_id_col] = id_values

    assert not pd.isnull(df[col_info.event_id_col]).any(), (
        data_type,
        [c for c in info_cols if pd.isnull(df[c]).any()],
        pd.isnull(df[col_info.event_id_col]).sum(),
        len(df.index),
        len_before,
        df.loc[:10, col_info.event_id_col],
        df.loc[pd.isnull(df[col_info.event_id_col]), [c for c in info_cols] + [col_info.event_id_col]][:10],
    )

    for column in temp_cols:
        del df[column]

    # Checking if there is only the one new column in the data frame now: The event_id_col.
    assert len(df.columns) == num_cols + 1, (len(df.columns), num_cols + 1)


def decompose_event_id(
    df: pd.DataFrame,
    remove_original_column: bool = False,
) -> None:
    col_info = UniqueEventIDColumnInfos()
    num_cols = len(df.columns)

    assert col_info.event_id_col in df.columns, col_info.data_type_col

    assert col_info.data_type_col not in df.columns, col_info.data_type_col
    assert col_info.exp_col not in df.columns, col_info.exp_col
    assert col_info.run_col not in df.columns, col_info.run_col
    assert col_info.evt_col not in df.columns, col_info.evt_col
    assert col_info.other_col not in df.columns, col_info.other_col

    split_id = df.loc[:, col_info.event_id_col].str.rsplit(pat=col_info.split_str, n=col_info.split_n, expand=True)
    assert not pd.isnull(split_id).any().any()

    df.loc[:, col_info.data_type_col] = split_id[0].astype("category")
    df.loc[:, col_info.exp_col] = split_id[1].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.run_col] = split_id[2].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.evt_col] = split_id[3].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.other_col] = split_id[4].astype("float")

    if remove_original_column:
        del df[col_info.event_id_col]
        assert len(df.columns) == num_cols + 4, (len(df.columns), num_cols)
    else:
        assert len(df.columns) == num_cols + 5, (len(df.columns), num_cols)


def recreate_event_id(df: pd.DataFrame, remove_original_columns: bool = False) -> None:
    col_info = UniqueEventIDColumnInfos()

    assert col_info.data_type_col in df.columns, col_info.data_type_col
    assert col_info.exp_col in df.columns, col_info.exp_col
    assert col_info.run_col in df.columns, col_info.run_col
    assert col_info.evt_col in df.columns, col_info.evt_col
    assert col_info.other_col in df.columns, col_info.other_col

    assert col_info.event_id_col not in df.columns, col_info.data_type_col

    info_cols = [col_info.data_type_col]
    int_info_cols = [col_info.exp_col, col_info.run_col, col_info.evt_col]
    float_info_cols = [col_info.other_col]
    original_info_cols = info_cols + int_info_cols + float_info_cols
    converted_cols = []

    for col in int_info_cols:
        converted_col_name = f"temp_{col}"
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, converted_col_name] = df[col].astype(float).astype(str)
        converted_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    for col in float_info_cols:
        rounded_col_name = f"temp_{col}_rounded"
        converted_col_name = f"temp_{col}_str"
        assert rounded_col_name not in df.columns, rounded_col_name
        assert converted_col_name not in df.columns, converted_col_name
        df.loc[:, rounded_col_name] = df[col].round(decimals=col_info.round_to)
        converted_cols.append(rounded_col_name)
        df.loc[:, converted_col_name] = df[rounded_col_name].astype(str)
        converted_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    df.loc[:, col_info.event_id_col] = pd.Series(df.loc[:, info_cols].values.tolist()).str.join(col_info.split_str)
    assert not pd.isnull(df[col_info.event_id_col]).any()

    for column in converted_cols:
        del df[column]
    if remove_original_columns:
        for column in original_info_cols:
            del df[column]


def add_unique_candidate_id(
    df: pd.DataFrame,
    id_column_name: str,
    event_id_name: str,
    candidate_id_name: str,
    check_if_unique: bool = True,
    mask: Optional[pd.Series] = None,
) -> np.ndarray:
    assert event_id_name in df.columns, event_id_name
    assert candidate_id_name in df.columns, candidate_id_name
    assert id_column_name not in df.columns, id_column_name

    active_mask = False
    if mask is None:
        mask = pd.Series(True, index=df.index)
    else:
        active_mask = True
        assert len(mask) == len(df.index), (len(mask), len(df.index))

    cand_id_col_name = candidate_id_name
    if df[candidate_id_name].dtype == np.float or df[candidate_id_name].dtype == np.int:
        temp_cand_id_as_str_name = f"temp_{candidate_id_name}"
        assert temp_cand_id_as_str_name not in df.columns, temp_cand_id_as_str_name

        df.loc[mask, temp_cand_id_as_str_name] = df[candidate_id_name].astype(str)
        if active_mask:
            df.loc[~mask, temp_cand_id_as_str_name] = "not_in_mask"

        cand_id_col_name = temp_cand_id_as_str_name

    cols_to_combine = [event_id_name, cand_id_col_name]
    df.loc[mask, id_column_name] = pd.Series(df.loc[mask, cols_to_combine].values.tolist()).str.join("_")
    if active_mask:
        df.loc[~mask, id_column_name] = "not_in_mask"

    assert not pd.isnull(df.loc[mask, id_column_name]).any()

    assert id_column_name in df.columns, id_column_name
    if check_if_unique:
        assert len(df.loc[mask].index) == len(df.loc[mask, id_column_name].unique()), (
            len(df.loc[mask].index),
            len(df.loc[mask, id_column_name].unique()),
        )

    if cand_id_col_name != candidate_id_name:
        del df[cand_id_col_name]

    return df.loc[mask, id_column_name].values
