"""
Steering file for the n-tuple file generation for the B -> D(*) tau nu analysis.
Felix Metzner 2018
"""

import basf2

from modularAnalysis import setAnalysisConfigParams, statistics

from analysistools.processing import job_interface
from analysistools.processing import path_utilities
from analysistools.processing import reco_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import b2bii_utils
from rdstar.online_analysis.decay_utilities import Reconstructor
from rdstar.online_analysis import observables, decay_utilities, skim_cuts, decay_channels


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(
    configParametersAndValues={"mcMatchingVersion": "Belle"},
    path=main_path,
)


# Get Final State Particles
main_path.add_module(module="CurlerCloneMarker")
decay_utilities.fsp.fsp_collection.get_final_state_particles(
    path=main_path,
    use_mc_info=cmd_args.config["use_mc_info"],
)
skim_cuts_utils.apply_skim_from_dict(
    path=main_path,
    skim_dict=skim_cuts.pre_skim_cuts,
)

reconstructor = Reconstructor(
    path=main_path,
    config=cmd_args.config,
    decay_channels_module=decay_channels,
)

reconstructor.reconstruct_tau_for_gen()


# Build intermediate D mesons
reconstructor.reconstruct_D()
reconstructor.reconstruct_Dstar()
reconstructor.reconstruct_D2star_for_gen()


# Build signal side B meson
reconstructor.reconstruct_sig_B()
skim_cuts_utils.apply_skim_from_dict(
    path=main_path,
    skim_dict=skim_cuts.post_signal_skim_cuts,
)


# Get tag side B meson via FEI
reco_utilities.reconstruct_tag_B(
    path=main_path,
    config=cmd_args.config,
    b_tag_chl_dict=decay_channels.BTag_decay_channels,
    decay_chl_dict=decay_channels.decay_channel_dict,
)
skim_cuts_utils.apply_skim_from_dict(
    path=main_path,
    skim_dict=skim_cuts.post_tag_skim_cuts,
)


# Build Upsilon(4S)
reconstructor.reconstruct_upsilon4s()
skim_cuts_utils.apply_skim_from_dict(
    path=main_path,
    skim_dict=skim_cuts.post_reco_skim_cuts,
)


##########
# Output #
##########

# # Get Decay Tree info if MC info is available
# If this should be used again: Set use_decay_hash to True in UpsilonVariables in _create_reco_level_collection!
# and add the respective output files to the output info in the batch configs!
# if config["use_mc_info"] is True:
#     main_path.add_module(
#         module="ParticleMCDecayString",
#         fileName=(cmd_args.output_dict["Y4S_decay_trees"]),
#         listName="Upsilon(4S):all"
#     )


up4s_coll = observables.get_upsilon4s_collection(configuration=cmd_args.config)
main_path.add_module(
    "VariablesToNtuple",
    fileName=cmd_args.output_dict["Y4S_ntuple"],
    particleList="Upsilon(4S):all",
    variables=up4s_coll.as_list(),
)


# if config["use_mc_info"] is True:
# If this should be used again: Add the respective output files to the output info in the batch configs!
#     # Move this VariablesToNtuple call in front of the last skim, if all
#     # possible Upsilon(4S) candidates shall be written out...
#     ups4s_gen_coll = observables.get_gen_level_upsilon4s_vars(configuration=cmd_args.config)
#     main_path.add_module(
#         module="VariablesToNtuple",
#         fileName=cmd_args.output_dict["Y4S_generator_level_reco_ntuple"],
#         particleList="Upsilon(4S):all_gen",
#         variables=ups4s_gen_coll.as_list(),
#     )
#
#     b_gen_coll = observables.get_gen_level_b_collection(configuration=cmd_args.config)
#     b_direct_gen_coll = observables.get_direct_gen_level_b_collection(configuration=cmd_args.config)
#     main_path.add_module(
#         module="VariablesToNtuple",
#         fileName=cmd_args.output_dict["B_minus_generator_info_ntuple"],
#         particleList="B-:sig_gen",
#         variables=b_gen_coll.as_list(),
#     )
#     main_path.add_module(
#         module="VariablesToNtuple",
#         fileName=cmd_args.output_dict["B_zero_generator_info_ntuple"],
#         particleList="B0:sig_gen",
#         variables=b_gen_coll.as_list(),
#     )
#
#     fillParticleListsFromMC([("B-:mc", "")], addDaughters=True, path=main_path)
#     fillParticleListsFromMC([("B0:mc", "")], addDaughters=True, path=main_path)
#     main_path.add_module(
#         module="VariablesToNtuple",
#         fileName=cmd_args.output_dict["B_minus_generator_direct_info_ntuple"],
#         particleList="B-:mc",
#         variables=b_direct_gen_coll.as_list(),
#     )
#     main_path.add_module(
#         module="VariablesToNtuple",
#         fileName=cmd_args.output_dict["B_zero_generator_direct_info_ntuple"],
#         particleList="B0:mc",
#         variables=b_direct_gen_coll.as_list(),
#     )

basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
